from django.contrib import admin
from django.db import models
from django.db.models import Count, Q
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from mptt.admin import DraggableMPTTAdmin
from parler.admin import TranslatableAdmin
from unfold.admin import ModelAdmin, TabularInline
from unfold.contrib.filters.admin import (
    DropdownFilter,
    RangeDateTimeFilter,
    RangeNumericListFilter,
    RelatedDropdownFilter,
    SliderNumericFilter,
)
from unfold.contrib.forms.widgets import WysiwygWidget
from unfold.decorators import action
from unfold.enums import ActionVariant

from blog.models.author import BlogAuthor
from blog.models.category import BlogCategory
from blog.models.comment import BlogComment
from blog.models.post import BlogPost, BlogPostTranslation
from blog.models.tag import BlogTag


class LikesCountFilter(RangeNumericListFilter):
    title = _("Likes")
    parameter_name = "likes_count"

    def queryset(self, request, queryset):
        filters = {}

        queryset = queryset.annotate(
            likes_count_annotation=Count("likes", distinct=True)
        )

        value_from = self.used_parameters.get(f"{self.parameter_name}_from")
        if value_from and value_from != "":
            filters["likes_count_annotation__gte"] = value_from

        value_to = self.used_parameters.get(f"{self.parameter_name}_to")
        if value_to and value_to != "":
            filters["likes_count_annotation__lte"] = value_to

        return queryset.filter(**filters) if filters else queryset

    def expected_parameters(self):
        return [
            f"{self.parameter_name}_from",
            f"{self.parameter_name}_to",
        ]


class CommentsCountFilter(RangeNumericListFilter):
    title = _("Comments")
    parameter_name = "comments_count"

    def queryset(self, request, queryset):
        filters = {}

        queryset = queryset.annotate(
            comments_count_annotation=Count(
                "comments", distinct=True, filter=Q(comments__approved=True)
            )
        )

        value_from = self.used_parameters.get(f"{self.parameter_name}_from")
        if value_from and value_from != "":
            filters["comments_count_annotation__gte"] = value_from

        value_to = self.used_parameters.get(f"{self.parameter_name}_to")
        if value_to and value_to != "":
            filters["comments_count_annotation__lte"] = value_to

        return queryset.filter(**filters) if filters else queryset

    def expected_parameters(self):
        return [
            f"{self.parameter_name}_from",
            f"{self.parameter_name}_to",
        ]


class TagsCountFilter(RangeNumericListFilter):
    title = _("Tags")
    parameter_name = "tags_count"

    def queryset(self, request, queryset):
        filters = {}

        queryset = queryset.annotate(
            tags_count_annotation=Count(
                "tags", distinct=True, filter=Q(tags__active=True)
            )
        )

        value_from = self.used_parameters.get(f"{self.parameter_name}_from")
        if value_from and value_from != "":
            filters["tags_count_annotation__gte"] = value_from

        value_to = self.used_parameters.get(f"{self.parameter_name}_to")
        if value_to and value_to != "":
            filters["tags_count_annotation__lte"] = value_to

        return queryset.filter(**filters) if filters else queryset

    def expected_parameters(self):
        return [
            f"{self.parameter_name}_from",
            f"{self.parameter_name}_to",
        ]


class PostsCountFilter(RangeNumericListFilter):
    title = _("Posts")
    parameter_name = "posts_count"

    def queryset(self, request, queryset):
        filters = {}

        if hasattr(queryset.model, "blog_posts"):
            queryset = queryset.annotate(
                posts_count_annotation=Count("blog_posts", distinct=True)
            )
        else:
            queryset = queryset.annotate(
                posts_count_annotation=Count("posts", distinct=True)
            )

        value_from = self.used_parameters.get(f"{self.parameter_name}_from")
        if value_from and value_from != "":
            filters["posts_count_annotation__gte"] = value_from

        value_to = self.used_parameters.get(f"{self.parameter_name}_to")
        if value_to and value_to != "":
            filters["posts_count_annotation__lte"] = value_to

        return queryset.filter(**filters) if filters else queryset

    def expected_parameters(self):
        return [
            f"{self.parameter_name}_from",
            f"{self.parameter_name}_to",
        ]


class PublishStatusFilter(DropdownFilter):
    title = _("Publish Status")
    parameter_name = "publish_status"

    def lookups(self, request, model_admin):
        return [
            ("published", _("Published")),
            ("draft", _("Draft")),
            ("scheduled", _("Scheduled")),
        ]

    def queryset(self, request, queryset):
        if self.value() == "published":
            return queryset.filter(
                is_published=True, published_at__lte=timezone.now()
            )
        elif self.value() == "draft":
            return queryset.filter(is_published=False)
        elif self.value() == "scheduled":
            return queryset.filter(
                is_published=True, published_at__gt=timezone.now()
            )
        return queryset


class BlogPostTranslationInline(TabularInline):
    model = BlogPostTranslation
    extra = 0
    fields = ("language_code", "title", "subtitle")
    show_change_link = True

    tab = True


class BlogCommentInline(TabularInline):
    model = BlogComment
    extra = 0
    fields = ("user", "content_preview", "approved", "created_at")
    readonly_fields = ("content_preview", "created_at")
    show_change_link = True

    tab = True

    def content_preview(self, obj):
        content = (
            obj.safe_translation_getter("content", any_language=True) or ""
        )
        return content[:50] + "..." if len(content) > 50 else content

    content_preview.short_description = _("Content Preview")


@admin.register(BlogAuthor)
class BlogAuthorAdmin(ModelAdmin, TranslatableAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = False
    list_filter_submit = True
    list_filter_sheet = True

    list_display = (
        "user_display",
        "bio_preview",
        "posts_count",
        "total_likes_display",
        "website_link",
    )
    search_fields = (
        "user__email",
        "user__username",
        "user__first_name",
        "user__last_name",
        "translations__bio",
    )
    list_select_related = ["user"]
    readonly_fields = ["id", "total_likes_display", "posts_count"]

    fieldsets = (
        (
            _("Basic Information"),
            {
                "fields": ("user", "website"),
                "classes": ("wide",),
            },
        ),
        (
            _("Biography"),
            {
                "fields": ("bio",),
                "classes": ("wide",),
            },
        ),
        (
            _("Statistics"),
            {
                "fields": ("posts_count", "total_likes_display"),
                "classes": ("collapse",),
            },
        ),
    )

    def user_display(self, obj):
        return format_html(
            '<div class="flex items-center gap-2">'
            '<strong class="text-base-900 dark:text-base-100">{}</strong>'
            '<span class="text-base-500 dark:text-base-400">({}</span>'
            "</div>",
            obj.user.full_name or obj.user.username,
            obj.user.email,
        )

    user_display.short_description = _("User")

    def bio_preview(self, obj):
        bio = obj.safe_translation_getter("bio", any_language=True) or ""
        if len(bio) > 50:
            return f"{bio[:50]}..."
        return bio

    bio_preview.short_description = _("Bio")

    def posts_count(self, obj):
        count = obj.blog_posts.count()
        return format_html(
            '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300 rounded-full">{}</span>',
            count,
        )

    posts_count.short_description = _("Posts")

    def total_likes_display(self, obj):
        total = obj.total_likes_received
        return format_html(
            '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300 rounded-full gap-1">'
            "<span>❤️</span>"
            "<span>{}</span>"
            "</span>",
            total,
        )

    total_likes_display.short_description = _("Total Likes")

    def website_link(self, obj):
        if obj.website:
            return format_html(
                '<a href="{}" target="_blank" class="inline-flex items-center gap-1 text-blue-600 dark:text-blue-400 hover:text-blue-800 dark:hover:text-blue-300">'
                "<span>🔗</span><span>Website</span>"
                "</a>",
                obj.website,
            )
        return "-"

    website_link.short_description = _("Website")


@admin.register(BlogTag)
class BlogTagAdmin(ModelAdmin, TranslatableAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = False
    list_filter_submit = True
    list_filter_sheet = True

    list_display = (
        "name_display",
        "active_badge",
        "active",
        "posts_count_badge",
        "sort_order",
    )
    list_filter = ("active", PostsCountFilter)
    list_editable = ("active",)
    search_fields = ("translations__name",)
    ordering = ("sort_order",)
    readonly_fields = [
        "sort_order",
    ]

    fieldsets = (
        (
            _("Tag Information"),
            {
                "fields": ("name", "active"),
                "classes": ("wide",),
            },
        ),
        (
            _("Organization"),
            {
                "fields": ("sort_order",),
                "classes": ("wide",),
            },
        ),
    )

    def name_display(self, obj):
        name = (
            obj.safe_translation_getter("name", any_language=True)
            or "Unnamed Tag"
        )
        return format_html(
            '<strong class="text-base-900 dark:text-base-100">{}</strong>', name
        )

    name_display.short_description = _("Name")

    def active_badge(self, obj):
        if obj.active:
            return format_html(
                '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300 rounded-full gap-1">'
                "<span>✓</span>"
                "<span>Active</span>"
                "</span>"
            )
        return format_html(
            '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300 rounded-full gap-1">'
            "<span>✗</span>"
            "<span>Inactive</span>"
            "</span>"
        )

    active_badge.short_description = _("Status")

    def posts_count_badge(self, obj):
        count = obj.blog_posts.count()
        return format_html(
            '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300 rounded-full">{}</span>',
            count,
        )

    posts_count_badge.short_description = _("Posts")


@admin.register(BlogCategory)
class BlogCategoryAdmin(ModelAdmin, TranslatableAdmin, DraggableMPTTAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = False

    mptt_indent_field = "translations__name"
    list_per_page = 15
    list_display = (
        "tree_actions",
        "indented_title",
        "category_image",
        "posts_count_display",
        "recursive_posts_display",
    )
    list_display_links = ("indented_title",)
    search_fields = ("translations__name", "translations__description")
    readonly_fields = [
        "id",
        "posts_count_display",
        "recursive_posts_display",
        "sort_order",
    ]

    fieldsets = (
        (
            _("Category Information"),
            {
                "fields": ("name", "description", "parent"),
                "classes": ("wide",),
            },
        ),
        (
            _("URL"),
            {
                "fields": ("slug",),
                "classes": ("wide",),
            },
        ),
        (
            _("Media"),
            {
                "fields": ("image",),
                "classes": ("wide",),
            },
        ),
        (
            _("Organization"),
            {
                "fields": ("sort_order",),
                "classes": ("collapse",),
            },
        ),
        (
            _("Statistics"),
            {
                "fields": ("posts_count_display", "recursive_posts_display"),
                "classes": ("collapse",),
            },
        ),
    )

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        qs = BlogCategory.objects.add_related_count(
            qs, BlogPost, "category", "posts_cumulative_count", cumulative=True
        )
        qs = BlogCategory.objects.add_related_count(
            qs, BlogPost, "category", "posts_count", cumulative=False
        )
        return qs

    def get_prepopulated_fields(self, request, obj=None):
        return {
            "slug": ("name",),
        }

    def category_image(self, obj):
        if obj.image:
            return format_html(
                '<img src="{}" style="max-height: 40px; max-width: 80px; border-radius: 4px;" />',
                obj.image.url,
            )
        return format_html(
            '<span class="text-base-400 dark:text-base-500">No image</span>'
        )

    category_image.short_description = _("Image")

    def posts_count_display(self, instance):
        count = getattr(instance, "posts_count", 0)
        return format_html(
            '<span class="inline-flex items-center px-2 py-1 text-xs font-semibold bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300 rounded-full">{}</span>',
            count,
        )

    posts_count_display.short_description = _("Direct Posts")

    def recursive_posts_display(self, instance):
        count = getattr(instance, "posts_cumulative_count", 0)
        return format_html(
            '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-purple-50 dark:bg-purple-900 text-purple-700 dark:text-purple-300 rounded-full">{}</span>',
            count,
        )

    recursive_posts_display.short_description = _("Total Posts (Tree)")


@admin.register(BlogPost)
class BlogPostAdmin(ModelAdmin, TranslatableAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = True
    list_horizontal_scrollbar_top = False

    formfield_overrides = {
        models.TextField: {
            "widget": WysiwygWidget,
        },
    }

    list_display = (
        "title_display",
        "category_badge",
        "author_display",
        "featured_badge",
        "featured",
        "is_published",
        "publish_status_badge",
        "engagement_metrics",
        "published_at",
        "image_preview",
    )
    list_filter = (
        "featured",
        PublishStatusFilter,
        ("published_at", RangeDateTimeFilter),
        ("created_at", RangeDateTimeFilter),
        ("category", RelatedDropdownFilter),
        ("author", RelatedDropdownFilter),
        ("tags", RelatedDropdownFilter),
        ("view_count", SliderNumericFilter),
        LikesCountFilter,
        CommentsCountFilter,
        TagsCountFilter,
    )
    list_editable = (
        "featured",
        "is_published",
    )
    search_fields = (
        "translations__title",
        "translations__subtitle",
        "slug",
        "translations__body",
        "author__user__email",
        "author__user__username",
    )
    autocomplete_fields = ["category", "author", "tags"]
    readonly_fields = [
        "id",
        "view_count",
        "created_at",
        "updated_at",
        "engagement_metrics",
        "seo_score",
    ]
    filter_horizontal = ["tags"]
    actions = [
        "mark_as_featured",
        "unmark_as_featured",
        "publish_posts",
        "unpublish_posts",
        "increment_view_count",
        "reset_view_count",
    ]
    date_hierarchy = "published_at"
    save_on_top = True
    list_per_page = 25
    list_select_related = ["category", "author", "author__user"]

    fieldsets = (
        (
            _("Content"),
            {
                "fields": ("title", "subtitle", "body"),
                "classes": ("wide",),
            },
        ),
        (
            _("Media"),
            {
                "fields": ("image",),
                "classes": ("wide",),
            },
        ),
        (
            _("Organization"),
            {
                "fields": ("slug", "category", "tags"),
                "classes": ("wide",),
            },
        ),
        (
            _("Publishing"),
            {
                "fields": (
                    "author",
                    "featured",
                    "is_published",
                    "published_at",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("Engagement"),
            {
                "fields": ("view_count", "engagement_metrics"),
                "classes": ("collapse",),
            },
        ),
        (
            _("SEO"),
            {
                "fields": ("seo_score",),
                "classes": ("collapse",),
            },
        ),
        (
            _("System"),
            {
                "fields": ("id", "created_at", "updated_at"),
                "classes": ("collapse",),
            },
        ),
    )

    inlines = [BlogPostTranslationInline, BlogCommentInline]

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .with_likes_count_annotation()
            .with_comments_count_annotation()
            .with_tags_count_annotation()
        )

    def title_display(self, obj):
        title = (
            obj.safe_translation_getter("title", any_language=True)
            or "Untitled"
        )
        return format_html(
            '<strong class="text-base-900 dark:text-base-100">{}</strong>',
            title[:50],
        )

    title_display.short_description = _("Title")

    def category_badge(self, obj):
        if obj.category:
            category_name = (
                obj.category.safe_translation_getter("name", any_language=True)
                or "Unnamed"
            )
            return format_html(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium bg-indigo-50 dark:bg-indigo-900 text-indigo-700 dark:text-indigo-300 rounded-full">{}</span>',
                category_name,
            )
        return format_html(
            '<span class="text-base-400 dark:text-base-500">No category</span>'
        )

    category_badge.short_description = _("Category")

    def author_display(self, obj):
        if obj.author:
            author_name = obj.author.user.full_name or obj.author.user.username
            return format_html(
                '<span class="font-medium text-base-700 dark:text-base-300">{}</span>',
                author_name,
            )
        return format_html(
            '<span class="text-base-400 dark:text-base-500">No author</span>'
        )

    author_display.short_description = _("Author")

    def featured_badge(self, obj):
        if obj.featured:
            return format_html(
                '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-yellow-50 dark:bg-yellow-900 text-yellow-700 dark:text-yellow-300 rounded-full gap-1">'
                "<span>⭐</span>"
                "<span>Featured</span>"
                "</span>"
            )
        return ""

    featured_badge.short_description = _("Featured")

    def publish_status_badge(self, obj):
        if obj.is_published and obj.published_at:
            if obj.published_at <= timezone.now():
                return format_html(
                    '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300 rounded-full gap-1">'
                    "<span>✓</span>"
                    "<span>Published</span>"
                    "</span>"
                )
            else:
                return format_html(
                    '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-300 rounded-full gap-1">'
                    "<span>📅</span>"
                    "<span>Scheduled</span>"
                    "</span>"
                )
        return format_html(
            '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-orange-50 dark:bg-orange-900 text-orange-700 dark:text-orange-300 rounded-full gap-1">'
            "<span>📝</span>"
            "<span>Draft</span>"
            "</span>"
        )

    publish_status_badge.short_description = _("Status")

    def engagement_metrics(self, obj):
        likes = obj.likes_count
        comments = obj.comments_count
        views = obj.view_count

        return format_html(
            '<div class="text-sm text-base-700 dark:text-base-300 flex items-center gap-3">'
            '<span class="flex items-center gap-1 text-red-600 dark:text-red-400">'
            "<span>❤️</span><span>{}</span>"
            "</span>"
            '<span class="flex items-center gap-1 text-blue-600 dark:text-blue-400">'
            "<span>💬</span><span>{}</span>"
            "</span>"
            '<span class="flex items-center gap-1 text-green-600 dark:text-green-400">'
            "<span>👀</span><span>{}</span>"
            "</span>"
            "</div>",
            likes,
            comments,
            views,
        )

    engagement_metrics.short_description = _("Engagement")

    def seo_score(self, obj):
        score = 0
        title = obj.safe_translation_getter("title", any_language=True) or ""
        subtitle = (
            obj.safe_translation_getter("subtitle", any_language=True) or ""
        )

        if len(title) > 10 and len(title) < 60:
            score += 25
        if subtitle:
            score += 25
        if obj.image:
            score += 25
        if obj.tags.exists():
            score += 25

        if score >= 75:
            bg_class = "bg-green-50 dark:bg-green-900"
            text_class = "text-green-700 dark:text-green-300"
        elif score >= 50:
            bg_class = "bg-yellow-50 dark:bg-yellow-900"
            text_class = "text-yellow-700 dark:text-yellow-300"
        else:
            bg_class = "bg-red-50 dark:bg-red-900"
            text_class = "text-red-700 dark:text-red-300"

        return format_html(
            '<span class="inline-flex items-center px-2 py-1 text-xs font-medium {} {} rounded-full">{}%</span>',
            bg_class,
            text_class,
            score,
        )

    seo_score.short_description = _("SEO Score")

    def image_preview(self, obj):
        if obj.image:
            return format_html(
                '<img src="{}" style="max-height: 100px; max-width: 80px; border-radius: 4px; object-fit: cover;" />',
                obj.image.url,
            )
        return format_html(
            '<span class="text-base-400 dark:text-base-500">No image</span>'
        )

    image_preview.short_description = ""

    def get_prepopulated_fields(self, request, obj=None):
        return {
            "slug": ("title",),
        }

    @action(
        description=_("Mark selected posts as featured"),
        variant=ActionVariant.PRIMARY,
        icon="star",
    )
    def mark_as_featured(self, request, queryset):
        updated = queryset.update(featured=True)
        self.message_user(
            request,
            _("%(count)d posts were successfully marked as featured.")
            % {"count": updated},
        )

    @action(
        description=_("Remove featured mark from selected posts"),
        variant=ActionVariant.WARNING,
        icon="star_border",
    )
    def unmark_as_featured(self, request, queryset):
        updated = queryset.update(featured=False)
        self.message_user(
            request,
            _("%(count)d posts were successfully unmarked as featured.")
            % {"count": updated},
        )

    @action(
        description=_("Publish selected posts"),
        variant=ActionVariant.SUCCESS,
        icon="publish",
    )
    def publish_posts(self, request, queryset):
        updated = queryset.update(
            is_published=True, published_at=timezone.now()
        )
        self.message_user(
            request,
            _("%(count)d posts were successfully published.")
            % {"count": updated},
        )

    @action(
        description=_("Unpublish selected posts"),
        variant=ActionVariant.WARNING,
        icon="unpublished",
    )
    def unpublish_posts(self, request, queryset):
        updated = queryset.update(is_published=False)
        self.message_user(
            request,
            _("%(count)d posts were successfully unpublished.")
            % {"count": updated},
        )

    @action(
        description=_("Increment view count by 100"),
        variant=ActionVariant.INFO,
        icon="visibility",
    )
    def increment_view_count(self, request, queryset):
        for post in queryset:
            post.view_count += 100
            post.save(update_fields=["view_count"])

        self.message_user(
            request,
            _("View count increased by 100 for %(count)d posts.")
            % {"count": queryset.count()},
        )

    @action(
        description=_("Reset view count to zero"),
        variant=ActionVariant.WARNING,
        icon="visibility_off",
    )
    def reset_view_count(self, request, queryset):
        updated = queryset.update(view_count=0)
        self.message_user(
            request,
            _("View count reset to zero for %(count)d posts.")
            % {"count": updated},
        )


@admin.register(BlogComment)
class BlogCommentAdmin(ModelAdmin, TranslatableAdmin, DraggableMPTTAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = True

    mptt_indent_field = "user"
    list_per_page = 30
    list_display = (
        "tree_actions",
        "indented_title",
        "user_display",
        "post_link",
        "approval_badge",
        "approved",
        "engagement_display",
        "created_at",
    )
    list_display_links = ("indented_title",)
    date_hierarchy = "created_at"
    search_fields = (
        "user__email",
        "user__username",
        "user__first_name",
        "user__last_name",
        "post__translations__title",
        "post__translations__subtitle",
        "translations__content",
    )
    list_filter = (
        "approved",
        ("created_at", RangeDateTimeFilter),
        ("post", RelatedDropdownFilter),
        ("user", RelatedDropdownFilter),
    )
    list_select_related = ["post", "user", "parent"]
    list_editable = ("approved",)
    actions = [
        "approve_comments",
        "unapprove_comments",
        "mark_as_spam",
    ]

    fieldsets = (
        (
            _("Comment Content"),
            {
                "fields": ("content",),
                "classes": ("wide",),
            },
        ),
        (
            _("Relations"),
            {
                "fields": ("post", "user", "parent"),
                "classes": ("wide",),
            },
        ),
        (
            _("Moderation"),
            {
                "fields": ("approved",),
                "classes": ("wide",),
            },
        ),
        (
            _("Engagement"),
            {
                "fields": ("engagement_display",),
                "classes": ("collapse",),
            },
        ),
    )

    readonly_fields = ["engagement_display"]

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("post", "user", "parent")
        )

    def user_display(self, obj):
        if obj.user:
            full_name = obj.user.full_name
            username = obj.user.username
            email = obj.user.email

            display_name = full_name if full_name else username
            return format_html(
                '<div class="text-sm">'
                '<div class="font-medium text-base-900 dark:text-base-100">{}</div>'
                '<div class="text-base-500 dark:text-base-400">{}</div>'
                "</div>",
                display_name,
                email,
            )
        return format_html(
            '<span class="text-base-400 dark:text-base-500 italic">Anonymous</span>'
        )

    user_display.short_description = _("User")

    def post_link(self, obj):
        if obj.post:
            title = (
                obj.post.safe_translation_getter("title", any_language=True)
                or f"Post {obj.post.id}"
            )
            url = f"/admin/blog/blogpost/{obj.post.id}/change/"
            return format_html(
                '<a href="{}" class="text-blue-600 dark:text-blue-400 hover:text-blue-800 dark:hover:text-blue-300 font-medium">{}</a>',
                url,
                title[:30] + "..." if len(title) > 30 else title,
            )
        return format_html(
            '<span class="text-base-400 dark:text-base-500">No post</span>'
        )

    post_link.short_description = _("Post")

    def approval_badge(self, obj):
        if obj.approved:
            return format_html(
                '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-300 rounded-full gap-1">'
                "<span>✓</span>"
                "<span>Approved</span>"
                "</span>"
            )
        return format_html(
            '<span class="inline-flex items-center justify-center px-2 py-1 text-xs font-medium bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-300 rounded-full gap-1">'
            "<span>⏳</span>"
            "<span>Pending</span>"
            "</span>"
        )

    approval_badge.short_description = _("Status")

    def engagement_display(self, obj):
        likes = obj.likes_count
        replies = obj.replies_count

        return format_html(
            '<div class="text-sm text-base-700 dark:text-base-300 flex items-center gap-3">'
            '<span class="flex items-center gap-1 text-red-600 dark:text-red-400">'
            "<span>❤️</span><span>{}</span>"
            "</span>"
            '<span class="flex items-center gap-1 text-blue-600 dark:text-blue-400">'
            "<span>💬</span><span>{}</span>"
            "</span>"
            "</div>",
            likes,
            replies,
        )

    engagement_display.short_description = _("Engagement")

    @action(
        description=_("Approve selected comments"),
        variant=ActionVariant.SUCCESS,
        icon="check_circle",
    )
    def approve_comments(self, request, queryset):
        updated = queryset.update(approved=True)
        self.message_user(
            request,
            _("%(count)d comments were successfully approved.")
            % {"count": updated},
        )

    @action(
        description=_("Unapprove selected comments"),
        variant=ActionVariant.WARNING,
        icon="cancel",
    )
    def unapprove_comments(self, request, queryset):
        updated = queryset.update(approved=False)
        self.message_user(
            request,
            _("%(count)d comments were successfully unapproved.")
            % {"count": updated},
        )

    @action(
        description=_("Mark as spam and delete"),
        variant=ActionVariant.DANGER,
        icon="report",
    )
    def mark_as_spam(self, request, queryset):
        count = queryset.count()
        queryset.delete()
        self.message_user(
            request,
            _("%(count)d comments were marked as spam and deleted.")
            % {"count": count},
        )



from django.urls import path
from rest_framework.urlpatterns import format_suffix_patterns

from tag.views.tag import TagViewSet
from tag.views.tagged_item import TaggedItemViewSet

urlpatterns = [
    path(
        "tag",
        TagViewSet.as_view({"get": "list", "post": "create"}),
        name="tag-list",
    ),
    path(
        "tag/<int:pk>",
        TagViewSet.as_view(
            {
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            }
        ),
        name="tag-detail",
    ),
    path(
        "tagged-item",
        TaggedItemViewSet.as_view({"get": "list", "post": "create"}),
        name="tagged-item-list",
    ),
    path(
        "tagged-item/<int:pk>",
        TaggedItemViewSet.as_view(
            {
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            }
        ),
        name="tagged-item-detail",
    ),
]

urlpatterns = format_suffix_patterns(urlpatterns)



from django.db import models
from django.utils.translation import gettext_lazy as _


class NotificationKindEnum(models.TextChoices):
    ERROR = "ERROR", _("Error")
    SUCCESS = "SUCCESS", _("Success")
    INFO = "INFO", _("Info")
    WARNING = "WARNING", _("Warning")
    DANGER = "DANGER", _("Danger")

    @classmethod
    def get_alert_types(cls) -> list[str]:
        return [cls.ERROR, cls.WARNING, cls.DANGER]

    @classmethod
    def get_positive_types(cls) -> list[str]:
        return [cls.SUCCESS, cls.INFO]


class NotificationCategoryEnum(models.TextChoices):
    ORDER = "ORDER", _("Order")
    PAYMENT = "PAYMENT", _("Payment")
    SHIPPING = "SHIPPING", _("Shipping")
    CART = "CART", _("Cart")
    PRODUCT = "PRODUCT", _("Product")
    ACCOUNT = "ACCOUNT", _("Account")
    SECURITY = "SECURITY", _("Security")
    PROMOTION = "PROMOTION", _("Promotion")
    SYSTEM = "SYSTEM", _("System")
    REVIEW = "REVIEW", _("Review")
    WISHLIST = "WISHLIST", _("Wishlist")
    SUPPORT = "SUPPORT", _("Support")
    NEWSLETTER = "NEWSLETTER", _("Newsletter")
    RECOMMENDATION = "RECOMMENDATION", _("Recommendation")


class NotificationPriorityEnum(models.TextChoices):
    LOW = "LOW", _("Low Priority")
    NORMAL = "NORMAL", _("Normal Priority")
    HIGH = "HIGH", _("High Priority")
    URGENT = "URGENT", _("Urgent Priority")
    CRITICAL = "CRITICAL", _("Critical Priority")



import json
from decimal import Decimal
from unittest.mock import patch

from django.conf import settings
from django.urls import reverse
from djmoney.money import Money
from rest_framework import status
from rest_framework.test import APIClient, APITestCase

from country.factories import CountryFactory
from order.enum.status import OrderStatus
from order.factories.order import OrderFactory
from order.models.order import Order
from pay_way.factories import PayWayFactory
from product.factories.product import ProductFactory
from region.factories import RegionFactory
from user.factories.account import UserAccountFactory

User = UserAccountFactory._meta.model


class CheckoutAPITestCase(APITestCase):
    def setUp(self):
        self.client = APIClient()
        self.user = UserAccountFactory(num_addresses=0)
        self.country = CountryFactory(num_regions=0)
        self.region = RegionFactory(country=self.country)
        self.pay_way = PayWayFactory()

        self.product1 = ProductFactory.create(
            stock=20, num_images=0, num_reviews=0, active=True
        )
        self.product2 = ProductFactory.create(
            stock=15, num_images=0, num_reviews=0, active=True
        )

        self.currency = str(self.product1.price.currency)

        self.checkout_url = reverse("order-list")
        self.checkout_data = {
            "email": "customer@example.com",
            "first_name": "John",
            "last_name": "Doe",
            "phone": "+12025550195",
            "street": "Main Street",
            "street_number": "123",
            "city": "Testville",
            "zipcode": "12345",
            "country": self.country.alpha_2,
            "region": self.region.alpha,
            "pay_way": self.pay_way.id,
            "shipping_price": "10.00",
            "items": [
                {"product": self.product1.id, "quantity": 2},
                {"product": self.product2.id, "quantity": 1},
            ],
        }

    @patch("order.signals.order_created.send")
    def test_checkout_successful(self, mock_signal):
        response = self.client.post(
            self.checkout_url,
            data=json.dumps(self.checkout_data),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(Order.objects.count(), 1)

        self.product1.refresh_from_db()
        self.product2.refresh_from_db()

        self.assertEqual(self.product1.stock, 20 - 2)
        self.assertEqual(self.product2.stock, 15 - 1)

        self.assertTrue(mock_signal.called)

    def test_checkout_insufficient_stock(self):
        product_limited = ProductFactory.create(
            stock=5, num_images=0, num_reviews=0, active=True
        )

        data = self.checkout_data.copy()
        data["items"] = [{"product": product_limited.id, "quantity": 10}]

        response = self.client.post(
            self.checkout_url,
            data=json.dumps(data),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("items", response.data)

        product_limited.refresh_from_db()
        self.assertEqual(product_limited.stock, 5)

    def test_checkout_authenticated_user(self):
        self.client.force_authenticate(user=self.user)

        response = self.client.post(
            self.checkout_url,
            data=json.dumps(self.checkout_data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        order = Order.objects.first()
        self.assertEqual(order.user, self.user)

    def test_checkout_invalid_data(self):
        invalid_data = self.checkout_data.copy()
        invalid_data["email"] = "invalid-email"

        response = self.client.post(
            self.checkout_url,
            data=json.dumps(invalid_data),
            content_type="application/json",
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertIn("email", response.data)

        self.assertEqual(Order.objects.count(), 0)


class OrderViewSetTestCase(APITestCase):
    def setUp(self):
        self.client = APIClient()

        self.user = User.objects.create_user(
            username="testuser",
            email="test@example.com",
            password="testpassword",
        )

        self.admin_user = User.objects.create_user(
            username="adminuser",
            email="admin@example.com",
            password="adminpassword",
            is_staff=True,
            is_superuser=True,
        )

        self.order1 = OrderFactory(user=self.user)
        self.order2 = OrderFactory(user=self.user)
        self.order3 = OrderFactory()

        product = ProductFactory(stock=10)
        self.order1.items.create(
            product=product,
            price=Money(
                amount=Decimal("50.00"), currency=settings.DEFAULT_CURRENCY
            ),
            quantity=2,
        )

        self.orders_url = reverse("order-list")
        self.order1_url = reverse("order-detail", kwargs={"pk": self.order1.pk})
        self.order_uuid_url = reverse(
            "order-retrieve-by-uuid", kwargs={"uuid": str(self.order1.uuid)}
        )
        self.my_orders_url = reverse("order-my-orders")
        self.cancel_order_url = reverse(
            "order-cancel", kwargs={"pk": self.order1.pk}
        )
        self.add_tracking = reverse(
            "order-add-tracking", kwargs={"pk": self.order1.pk}
        )
        self.update_status_url = reverse(
            "order-update-status", kwargs={"pk": self.order1.pk}
        )

    def test_list_orders_unauthenticated(self):
        self.client.force_authenticate(user=None)
        response = self.client.get(self.orders_url)
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_list_orders_admin(self):
        self.client.force_authenticate(user=self.admin_user)

        response = self.client.get(self.orders_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(len(response.data["results"]), Order.objects.count())

    def test_retrieve_order_by_id(self):
        self.client.force_authenticate(user=self.admin_user)

        response = self.client.get(self.order1_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data["id"], self.order1.id)

    def test_retrieve_order_by_uuid(self):
        self.client.force_authenticate(user=self.admin_user)

        response = self.client.get(self.order_uuid_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.data["uuid"], str(self.order1.uuid))

    def test_my_orders(self):
        self.client.force_authenticate(user=self.user)

        response = self.client.get(self.my_orders_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertGreaterEqual(len(response.data["results"]), 1)

        order_ids = [order["id"] for order in response.data["results"]]
        self.assertNotIn(self.order3.id, order_ids)

    def test_cancel_order(self):
        self.client.force_authenticate(user=self.admin_user)

        self.order1.status = OrderStatus.PENDING.value
        self.order1.save()

        response = self.client.post(self.cancel_order_url)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.order1.refresh_from_db()
        self.assertEqual(self.order1.status, OrderStatus.CANCELED.value)

    def test_add_tracking(self):
        self.client.force_authenticate(user=self.admin_user)

        self.order1.status = OrderStatus.PROCESSING.value
        self.order1.save()

        tracking_data = {
            "tracking_number": "TRACK123456",
            "shipping_carrier": "FedEx",
        }

        response = self.client.post(
            self.add_tracking,
            data=json.dumps(tracking_data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.order1.refresh_from_db()
        self.assertEqual(self.order1.tracking_number, "TRACK123456")
        self.assertEqual(self.order1.shipping_carrier, "FedEx")

    def test_update_order_status(self):
        self.client.force_authenticate(user=self.admin_user)

        self.order1.status = OrderStatus.PENDING.value
        self.order1.save()

        status_data = {"status": OrderStatus.PROCESSING.value}

        response = self.client.post(
            self.update_status_url,
            data=json.dumps(status_data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.order1.refresh_from_db()
        self.assertEqual(self.order1.status, OrderStatus.PROCESSING.value)

    def test_update_order_status_invalid_transition(self):
        self.client.force_authenticate(user=self.admin_user)

        self.order1.status = OrderStatus.PENDING.value
        self.order1.save()

        status_data = {"status": OrderStatus.COMPLETED.value}

        response = self.client.post(
            self.update_status_url,
            data=json.dumps(status_data),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.order1.refresh_from_db()
        self.assertEqual(self.order1.status, OrderStatus.PENDING.value)



# Generated by Django 5.2.6 on 2025-09-29 22:54

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('order', '0016_alter_orderhistorytranslation_options_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='order',
            name='metadata',
            field=models.JSONField(blank=True, default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
        migrations.AddField(
            model_name='order',
            name='private_metadata',
            field=models.JSONField(blank=True, default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
    ]



# Generated by Django 5.2.2 on 2025-06-07 17:12

import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('notification', '0009_remove_notificationuser_notificatio_user_id_f8507b_idx_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='category',
            field=models.CharField(choices=[('order', 'Order'), ('payment', 'Payment'), ('shipping', 'Shipping'), ('cart', 'Cart'), ('product', 'Product'), ('account', 'Account'), ('security', 'Security'), ('promotion', 'Promotion'), ('system', 'System'), ('review', 'Review'), ('wishlist', 'Wishlist'), ('support', 'Support'), ('newsletter', 'Newsletter'), ('recommendation', 'Recommendation')], db_index=True, default='system', max_length=50, verbose_name='Category'),
        ),
        migrations.AddField(
            model_name='notification',
            name='notification_type',
            field=models.CharField(blank=True, help_text='Specific type of notification for categorization', max_length=100, verbose_name='Notification Type'),
        ),
        migrations.AddField(
            model_name='notification',
            name='priority',
            field=models.CharField(choices=[('low', 'Low Priority'), ('normal', 'Normal Priority'), ('high', 'High Priority'), ('urgent', 'Urgent Priority'), ('critical', 'Critical Priority')], db_index=True, default='normal', max_length=20, verbose_name='Priority'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['category'], name='notification_category_ix'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['priority'], name='notification_priority_ix'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['notification_type'], name='notification_type_ix'),
        ),
        migrations.AddIndex(
            model_name='notification',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['category', 'priority'], name='notification_cat_pri_ix'),
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase

from product.factories.category import ProductCategoryFactory
from product.factories.product import ProductFactory
from product.models.category import ProductCategory
from user.factories.account import UserAccountFactory
from vat.factories import VatFactory

languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]
default_language = settings.PARLER_DEFAULT_LANGUAGE_CODE
User = get_user_model()


class CategoryModelTestCase(TestCase):
    def setUp(self):
        self.user = UserAccountFactory(num_addresses=0)
        self.vat = VatFactory()
        self.category = ProductCategoryFactory()
        self.sub_category = ProductCategoryFactory(parent=self.category)

    def test_fields(self):
        self.assertIsNotNone(self.category.slug)
        self.assertIsNotNone(self.category.id)
        self.assertTrue(self.category.active)

    def test_str_representation_no_parent(self):
        category_name = self.category.safe_translation_getter("name") or ""
        category_str = str(self.category)

        self.assertEqual(category_str, category_name)

    def test_get_ordering_queryset_with_parent(self):
        self.assertEqual(ProductCategory.objects.count(), 2)
        self.assertIn(self.sub_category, ProductCategory.objects.all())
        self.assertIn(self.category, ProductCategory.objects.all())

        parent_queryset = self.category.get_ordering_queryset()

        self.assertIn(self.sub_category, parent_queryset)

        self.assertIn(self.category, parent_queryset)

    def test_get_ordering_queryset_without_parent(self):
        no_parent_category = ProductCategoryFactory(
            slug="no-parent-category",
        )

        parent_queryset = no_parent_category.get_ordering_queryset()

        self.assertIn(no_parent_category, parent_queryset)

        self.assertIn(self.sub_category, parent_queryset)
        self.assertIn(self.category, parent_queryset)

        for descendant in self.category.get_descendants(include_self=True):
            self.assertIn(descendant, parent_queryset)

    def test_recursive_product_count_no_products(self):
        count = self.category.recursive_product_count
        self.assertEqual(count, 0)

    def test_recursive_product_count_one_product(self):
        ProductFactory(
            category=self.category, vat=self.vat, num_images=0, num_reviews=0
        )

        count = self.category.recursive_product_count
        self.assertEqual(count, 1)

    def test_recursive_product_count_multiple_products(self):
        ProductFactory(category=self.category, num_images=0, num_reviews=0)
        ProductFactory(category=self.sub_category, num_images=0, num_reviews=0)

        count = self.category.recursive_product_count
        self.assertEqual(count, 2)



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from tag.factories.tag import TagFactory
from tag.factories.tagged_item import TaggedProductFactory
from tag.models.tagged_item import TaggedItem
from product.factories.product import ProductFactory


class TaggedItemModelTestCase(TestCase):
    def setUp(self):
        self.tag = TagFactory(active=True)
        self.product = ProductFactory()
        self.tagged_item = TaggedProductFactory(
            tag=self.tag, content_object=self.product
        )

    def test_fields(self):
        self.assertEqual(self.tagged_item.tag, self.tag)
        self.assertEqual(self.tagged_item.object_id, self.product.id)
        self.assertEqual(self.tagged_item.content_object, self.product)

        expected_content_type = ContentType.objects.get_for_model(self.product)
        self.assertEqual(self.tagged_item.content_type, expected_content_type)

    def test_content_object_generic_relation(self):
        self.assertIsInstance(
            self.tagged_item.content_object, type(self.product)
        )
        self.assertEqual(self.tagged_item.content_object.id, self.product.id)

    def test_manager_active_tags_method(self):
        inactive_tag = TagFactory(active=False)
        inactive_tagged_item = TaggedProductFactory(
            tag=inactive_tag, content_object=ProductFactory()
        )

        active_items = TaggedItem.objects.active_tags()

        self.assertIn(self.tagged_item, active_items)
        self.assertNotIn(inactive_tagged_item, active_items)

    def test_manager_get_tags_for_method(self):
        another_tag = TagFactory(active=True)
        TaggedProductFactory(tag=another_tag, content_object=self.product)

        product_tags = TaggedItem.objects.get_tags_for(
            type(self.product), self.product.id
        )

        self.assertEqual(product_tags.count(), 2)

    def test_uuid_field(self):
        self.assertIsNotNone(self.tagged_item.uuid)

        other_tagged_item = TaggedProductFactory(
            tag=self.tag, content_object=ProductFactory()
        )
        self.assertNotEqual(self.tagged_item.uuid, other_tagged_item.uuid)

    def test_timestamp_fields(self):
        self.assertIsNotNone(self.tagged_item.created_at)
        self.assertIsNotNone(self.tagged_item.updated_at)

    def test_cascade_delete_on_tag(self):
        tagged_item_id = self.tagged_item.id

        self.tag.delete()

        self.assertFalse(TaggedItem.objects.filter(id=tagged_item_id).exists())

    def test_cascade_delete_on_content_type(self):
        tagged_item_id = self.tagged_item.id
        content_type = self.tagged_item.content_type

        content_type.delete()

        self.assertFalse(TaggedItem.objects.filter(id=tagged_item_id).exists())


class TaggedModelTestCase(TestCase):
    def setUp(self):
        self.tag1 = TagFactory(active=True)
        self.tag2 = TagFactory(active=True)
        self.product = ProductFactory()

    def test_tag_ids_property(self):
        self.assertEqual(self.product.tag_ids, [])

        TaggedProductFactory(tag=self.tag1, content_object=self.product)
        TaggedProductFactory(tag=self.tag2, content_object=self.product)

        tag_ids = self.product.tag_ids
        self.assertIn(self.tag1.id, tag_ids)
        self.assertIn(self.tag2.id, tag_ids)
        self.assertEqual(len(tag_ids), 2)

    def test_get_tags_by_object_ids_class_method(self):
        product2 = ProductFactory()

        TaggedProductFactory(tag=self.tag1, content_object=self.product)
        TaggedProductFactory(tag=self.tag2, content_object=self.product)
        TaggedProductFactory(tag=self.tag1, content_object=product2)

        object_ids = [self.product.id, product2.id]
        tags = ProductFactory._meta.model.get_tags_by_object_ids(
            object_ids, ProductFactory._meta.model
        )

        self.assertIn(self.tag1, tags)
        self.assertIn(self.tag2, tags)

    def test_get_tags_for_object_method(self):
        TaggedProductFactory(tag=self.tag1, content_object=self.product)
        TaggedProductFactory(tag=self.tag2, content_object=self.product)

        tags = self.product.get_tags_for_object()

        self.assertIn(self.tag1, tags)
        self.assertIn(self.tag2, tags)
        self.assertEqual(tags.count(), 2)

    def test_add_tag_method(self):
        tagged_item = TaggedItem(tag=self.tag1)

        self.product.add_tag(tagged_item)

        self.assertEqual(tagged_item.content_object, self.product)
        self.assertEqual(tagged_item.object_id, self.product.id)

    def test_remove_tag_method(self):
        tagged_item = TaggedProductFactory(
            tag=self.tag1, content_object=self.product
        )
        tagged_item_id = tagged_item.id

        self.product.remove_tag(tagged_item)

        self.assertFalse(TaggedItem.objects.filter(id=tagged_item_id).exists())

    def test_clear_tags_method(self):
        TaggedProductFactory(tag=self.tag1, content_object=self.product)
        TaggedProductFactory(tag=self.tag2, content_object=self.product)

        self.assertEqual(self.product.tags.count(), 2)

        self.product.clear_tags()

        self.assertEqual(self.product.tags.count(), 0)



from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters

from core.filters.camel_case_filters import CamelCaseTimeStampFilterSet
from core.filters.core import UUIDFilterMixin
from notification.models.user import NotificationUser


class NotificationUserFilter(UUIDFilterMixin, CamelCaseTimeStampFilterSet):
    user = filters.NumberFilter(
        field_name="user__id", help_text=_("Filter by user ID")
    )
    notification = filters.NumberFilter(
        field_name="notification__id", help_text=_("Filter by notification ID")
    )
    seen = filters.BooleanFilter(
        field_name="seen", help_text=_("Filter by seen status")
    )

    user__email = filters.CharFilter(
        field_name="user__email",
        lookup_expr="icontains",
        help_text=_("Filter by user email (case-insensitive)"),
    )
    user__first_name = filters.CharFilter(
        field_name="user__first_name",
        lookup_expr="icontains",
        help_text=_("Filter by user first name (case-insensitive)"),
    )
    user__last_name = filters.CharFilter(
        field_name="user__last_name",
        lookup_expr="icontains",
        help_text=_("Filter by user last name (case-insensitive)"),
    )
    user__is_active = filters.BooleanFilter(
        field_name="user__is_active",
        help_text=_("Filter by user active status"),
    )
    user__is_staff = filters.BooleanFilter(
        field_name="user__is_staff",
        help_text=_("Filter by user staff status"),
    )

    seen_after = filters.DateTimeFilter(
        field_name="seen_at",
        lookup_expr="gte",
        help_text=_("Filter notifications seen after this date"),
    )
    seen_before = filters.DateTimeFilter(
        field_name="seen_at",
        lookup_expr="lte",
        help_text=_("Filter notifications seen before this date"),
    )
    has_seen_at = filters.BooleanFilter(
        method="filter_has_seen_at",
        help_text=_(
            "Filter notifications that have been seen (true) or not seen (false)"
        ),
    )

    notification_kind = filters.CharFilter(
        field_name="notification__kind",
        help_text=_("Filter by notification kind"),
    )
    notification__kind = filters.CharFilter(
        field_name="notification__kind",
        help_text=_("Filter by notification kind"),
    )
    notification__category = filters.CharFilter(
        field_name="notification__category",
        help_text=_("Filter by notification category"),
    )
    notification__priority = filters.CharFilter(
        field_name="notification__priority",
        help_text=_("Filter by notification priority"),
    )
    notification__type = filters.CharFilter(
        field_name="notification__notification_type",
        lookup_expr="icontains",
        help_text=_("Filter by notification type (case-insensitive)"),
    )
    notification__link = filters.CharFilter(
        field_name="notification__link",
        lookup_expr="icontains",
        help_text=_("Filter by notification link (case-insensitive)"),
    )

    notification__title = filters.CharFilter(
        field_name="notification__translations__title",
        lookup_expr="icontains",
        help_text=_("Filter by notification title (case-insensitive)"),
    )
    notification__message = filters.CharFilter(
        field_name="notification__translations__message",
        lookup_expr="icontains",
        help_text=_("Filter by notification message (case-insensitive)"),
    )

    notification__is_expired = filters.BooleanFilter(
        method="filter_notification_is_expired",
        help_text=_("Filter by notification expiry status"),
    )
    notification__expires_after = filters.DateTimeFilter(
        field_name="notification__expiry_date",
        lookup_expr="gte",
        help_text=_("Filter notifications expiring after this date"),
    )
    notification__expires_before = filters.DateTimeFilter(
        field_name="notification__expiry_date",
        lookup_expr="lte",
        help_text=_("Filter notifications expiring before this date"),
    )

    unseen_only = filters.BooleanFilter(
        method="filter_unseen_only",
        help_text=_("Filter only unseen notifications"),
    )
    seen_only = filters.BooleanFilter(
        method="filter_seen_only", help_text=_("Filter only seen notifications")
    )
    recent_notifications = filters.BooleanFilter(
        method="filter_recent_notifications",
        help_text=_("Filter notifications from the last 7 days"),
    )
    high_priority = filters.BooleanFilter(
        method="filter_high_priority",
        help_text=_("Filter high priority notifications"),
    )

    user_ids = filters.CharFilter(
        method="filter_user_ids",
        help_text=_("Filter by multiple user IDs (comma-separated)"),
    )
    notification_ids = filters.CharFilter(
        method="filter_notification_ids",
        help_text=_("Filter by multiple notification IDs (comma-separated)"),
    )

    class Meta:
        model = NotificationUser
        fields = {
            "created_at": ["gte", "lte", "date"],
            "updated_at": ["gte", "lte", "date"],
            "uuid": ["exact"],
            "id": ["exact", "in"],
            "seen": ["exact"],
            "seen_at": ["gte", "lte", "date"],
            "user": ["exact"],
            "notification": ["exact"],
        }

    def filter_has_seen_at(self, queryset, name, value):
        """Filter notifications based on whether they have been seen."""
        if value is True:
            return queryset.filter(seen_at__isnull=False)
        elif value is False:
            return queryset.filter(seen_at__isnull=True)
        return queryset

    def filter_notification_is_expired(self, queryset, name, value):
        """Filter notifications based on expiry status."""
        from django.utils import timezone

        now = timezone.now()
        if value is True:
            return queryset.filter(
                notification__expiry_date__isnull=False,
                notification__expiry_date__lt=now,
            )
        elif value is False:
            return queryset.filter(
                Q(notification__expiry_date__isnull=True)
                | Q(notification__expiry_date__gte=now)
            )
        return queryset

    def filter_unseen_only(self, queryset, name, value):
        """Filter only unseen notifications."""
        if value is True:
            return queryset.filter(seen=False)
        return queryset

    def filter_seen_only(self, queryset, name, value):
        """Filter only seen notifications."""
        if value is True:
            return queryset.filter(seen=True)
        return queryset

    def filter_recent_notifications(self, queryset, name, value):
        """Filter notifications from the last 7 days."""
        if value is True:
            from django.utils import timezone
            from datetime import timedelta

            seven_days_ago = timezone.now() - timedelta(days=7)
            return queryset.filter(notification__created_at__gte=seven_days_ago)
        return queryset

    def filter_high_priority(self, queryset, name, value):
        """Filter high priority notifications."""
        if value is True:
            from notification.enum import NotificationPriorityEnum

            return queryset.filter(
                notification__priority=NotificationPriorityEnum.HIGH
            )
        return queryset

    def filter_user_ids(self, queryset, name, value):
        """Filter by multiple user IDs."""
        if value:
            try:
                user_ids = [
                    int(id.strip()) for id in value.split(",") if id.strip()
                ]
                return queryset.filter(user__id__in=user_ids)
            except ValueError:
                return queryset.none()
        return queryset

    def filter_notification_ids(self, queryset, name, value):
        """Filter by multiple notification IDs."""
        if value:
            try:
                notification_ids = [
                    int(id.strip()) for id in value.split(",") if id.strip()
                ]
                return queryset.filter(notification__id__in=notification_ids)
            except ValueError:
                return queryset.none()
        return queryset



import logging
from typing import Any

from django.db.models import FloatField, Model
from django.forms import ChoiceField, Field
from django.utils.translation import gettext_lazy as _
from measurement.base import BidimensionalMeasure, MeasureBase

from core.forms.measurement import MeasurementFormField
from core.utils.measurement import get_measurement

logger = logging.getLogger(__name__)


class MeasurementTypeError(TypeError):
    pass


class MeasurementField(FloatField):
    description = "Easily store, retrieve, and convert python measures."
    empty_strings_allowed = False
    MEASURE_BASES = (
        BidimensionalMeasure,
        MeasureBase,
    )
    default_error_messages = {
        "invalid_type": _(
            "'%(value)s' (%(type_given)s) value"
            " must be of type %(type_wanted)s."
        ),
    }

    def __init__(
        self,
        verbose_name: str | None = None,
        name: str | None = None,
        measurement: type[MeasureBase | BidimensionalMeasure] | None = None,
        unit_choices: list[tuple[str, str]] | None = None,
        *args,
        **kwargs,
    ):
        if measurement is None or not issubclass(
            measurement, self.MEASURE_BASES
        ):
            raise MeasurementTypeError(
                "MeasurementField requires a measurement subclass of MeasureBase."
            )

        self.measurement = measurement
        self.widget_args = {
            "measurement": measurement,
            "unit_choices": unit_choices,
        }

        super().__init__(verbose_name, name, *args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["measurement"] = self.measurement
        return name, path, args, kwargs

    def get_prep_value(self, value):
        if value is None:
            return None

        elif isinstance(value, self.MEASURE_BASES):
            return float(value.standard)

        else:
            return super().get_prep_value(value)

    def get_default_unit(self):
        unit_choices = self.widget_args["unit_choices"]
        if unit_choices:
            return unit_choices[0][0]
        return self.measurement.STANDARD_UNIT

    def from_db_value(self, value: float | None, *args, **kwargs):
        if value is None:
            return None

        return get_measurement(
            measure=self.measurement,
            value=value,
            original_unit=self.get_default_unit(),
        )

    def value_to_string(self, obj: Model):
        value = self.value_from_object(obj)
        if not isinstance(value, self.MEASURE_BASES):
            return value
        return "{}:{}".format(value.value, value.unit)

    def deserialize_value_from_string(self, value_str: str):
        try:
            value_part, unit = value_str.split(":", 1)
            value_float = float(value_part)
            measure = get_measurement(
                self.measurement, value=value_float, unit=unit
            )
            return measure
        except ValueError as e:
            logger.error(f"Error deserializing measurement: {e}")
            return None

    def to_python(self, value: Any):
        if value is None or isinstance(value, self.MEASURE_BASES):
            return value
        elif isinstance(value, str):
            parsed = self.deserialize_value_from_string(value)
            if parsed is not None:
                return parsed
        value = super().to_python(value)

        return_unit = self.get_default_unit()

        msg = 'You assigned a {} instead of {} to {}.{}.{}, unit was guessed to be "{}".'.format(
            type(value).__name__,
            str(self.measurement.__name__),
            self.model.__module__,
            self.model.__name__,
            self.name,
            return_unit,
        )
        logger.warning(msg)
        return get_measurement(
            measure=self.measurement,
            value=value,
            unit=return_unit,
        )

    def formfield(
        self,
        form_class: type[Field] | None = None,
        choices_form_class: type[ChoiceField] | None = None,
        **kwargs: Any,
    ) -> Field | None:
        if form_class is None:
            form_class = MeasurementFormField

        kwargs.update(self.widget_args)

        return super().formfield(
            form_class=form_class,
            choices_form_class=choices_form_class,
            **kwargs,
        )



from io import StringIO
from unittest.mock import patch

from django.core.management import call_command

from core.management.commands.clear_cache import Command


class TestClearCacheCommand:
    def test_command_help_text(self):
        command = Command()
        assert command.help == "Clears cache"

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_clear_cache_success(self, mock_cache):
        mock_cache.clear.return_value = None

        out = StringIO()
        command = Command()
        command.stdout = out

        command.handle()

        mock_cache.clear.assert_called_once()
        output = out.getvalue()
        assert "Successfully cleared cache" in output

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_clear_cache_exception(self, mock_cache):
        mock_cache.clear.side_effect = Exception("Cache error")

        out = StringIO()
        command = Command()
        command.stdout = out

        command.handle()

        mock_cache.clear.assert_called_once()
        output = out.getvalue()
        assert "Error clearing cache: Cache error" in output

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_clear_cache_via_call_command(self, mock_cache):
        mock_cache.clear.return_value = None

        out = StringIO()
        call_command("clear_cache", stdout=out)

        mock_cache.clear.assert_called_once()
        output = out.getvalue()
        assert "Successfully cleared cache" in output

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_clear_cache_connection_error(self, mock_cache):
        mock_cache.clear.side_effect = ConnectionError(
            "Cannot connect to cache"
        )

        out = StringIO()
        command = Command()
        command.stdout = out

        command.handle()

        mock_cache.clear.assert_called_once()
        output = out.getvalue()
        assert "Error clearing cache: Cannot connect to cache" in output

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_clear_cache_runtime_error(self, mock_cache):
        mock_cache.clear.side_effect = RuntimeError("Runtime error occurred")

        out = StringIO()
        command = Command()
        command.stdout = out

        command.handle()

        mock_cache.clear.assert_called_once()
        output = out.getvalue()
        assert "Error clearing cache: Runtime error occurred" in output

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_command_with_args_and_kwargs(self, mock_cache):
        mock_cache.clear.return_value = None

        out = StringIO()
        command = Command()
        command.stdout = out

        command.handle("arg1", "arg2", option1="value1", option2="value2")

        mock_cache.clear.assert_called_once()
        output = out.getvalue()
        assert "Successfully cleared cache" in output

    def test_command_instance_attributes(self):
        command = Command()

        assert hasattr(command, "help")
        assert hasattr(command, "handle")
        assert command.help == "Clears cache"

    @patch("core.management.commands.clear_cache.cache_instance")
    def test_multiple_command_executions(self, mock_cache):
        mock_cache.clear.return_value = None

        command = Command()

        for _ in range(3):
            out = StringIO()
            command.stdout = out
            command.handle()

            output = out.getvalue()
            assert "Successfully cleared cache" in output

        assert mock_cache.clear.call_count == 3



import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.db import transaction
from django.utils.crypto import get_random_string
from django.utils.http import urlsafe_base64_decode
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import serializers, status
from rest_framework.decorators import action

from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView

from core.api.permissions import IsOwnerOrAdmin
from core.api.serializers import ErrorResponseSerializer, DetailSerializer
from core.api.views import BaseModelViewSet
from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from user.filters.subscription import (
    SubscriptionTopicFilter,
    UserSubscriptionFilter,
)
from user.models.subscription import SubscriptionTopic, UserSubscription
from user.serializers.subscription import (
    BulkSubscriptionSerializer,
    SubscriptionTopicDetailSerializer,
    SubscriptionTopicSerializer,
    SubscriptionTopicWriteSerializer,
    UserSubscriptionDetailSerializer,
    UserSubscriptionSerializer,
    UserSubscriptionWriteSerializer,
)
from user.utils.subscription import send_subscription_confirmation

logger = logging.getLogger(__name__)

User = get_user_model()

subscription_topic_req_serializers: RequestSerializersConfig = {
    "create": SubscriptionTopicWriteSerializer,
    "update": SubscriptionTopicWriteSerializer,
    "partial_update": SubscriptionTopicWriteSerializer,
    "subscribe": None,
    "unsubscribe": None,
}

subscription_topic_res_serializers: ResponseSerializersConfig = {
    "create": SubscriptionTopicDetailSerializer,
    "list": SubscriptionTopicSerializer,
    "retrieve": SubscriptionTopicDetailSerializer,
    "update": SubscriptionTopicDetailSerializer,
    "partial_update": SubscriptionTopicDetailSerializer,
    "my_subscriptions": SubscriptionTopicSerializer,
    "subscribe": UserSubscriptionSerializer,
    "unsubscribe": DetailSerializer,
}


@extend_schema_view(
    **create_schema_view_config(
        model_class=SubscriptionTopic,
        display_config={
            "tag": "Subscription Topics",
        },
        request_serializers=subscription_topic_req_serializers,
        response_serializers=subscription_topic_res_serializers,
    )
)
class SubscriptionTopicViewSet(BaseModelViewSet):
    queryset = SubscriptionTopic.objects.filter(is_active=True)
    request_serializers = subscription_topic_req_serializers
    response_serializers = subscription_topic_res_serializers
    permission_classes = [IsAuthenticated]
    filterset_class = SubscriptionTopicFilter
    ordering_fields = ["category", "created_at", "updated_at", "slug"]
    ordering = ["category"]
    search_fields = ["translations__name", "translations__description", "slug"]

    @extend_schema(
        operation_id="getMySubscriptionTopics",
        summary=_("Get my subscriptions"),
        description=_(
            "Get the current user's subscribed and available subscription topics."
        ),
        tags=["Subscription Topics"],
        responses={
            200: {
                "type": "object",
                "properties": {
                    "subscribed": {
                        "type": "array",
                        "items": {
                            "$ref": "#/components/schemas/SubscriptionTopic"
                        },
                    },
                    "available": {
                        "type": "array",
                        "items": {
                            "$ref": "#/components/schemas/SubscriptionTopic"
                        },
                    },
                },
            },
            401: ErrorResponseSerializer,
        },
    )
    @action(detail=False, methods=["GET"])
    def my_subscriptions(self, request):
        user = request.user

        subscribed_topics = SubscriptionTopic.objects.filter(
            subscribers__user=user,
            subscribers__status=UserSubscription.SubscriptionStatus.ACTIVE,
            is_active=True,
        ).distinct()

        available_topics = SubscriptionTopic.objects.filter(
            is_active=True
        ).exclude(id__in=subscribed_topics.values_list("id", flat=True))

        response_serializer_class = self.get_response_serializer()
        subscribed_data = response_serializer_class(
            subscribed_topics, many=True, context=self.get_serializer_context()
        ).data
        available_data = response_serializer_class(
            available_topics, many=True, context=self.get_serializer_context()
        ).data

        response_data = {
            "subscribed": subscribed_data,
            "available": available_data,
        }

        return Response(response_data)

    @extend_schema(
        operation_id="subscribeToTopic",
        summary=_("Subscribe to a topic"),
        description=_(
            "Subscribe the current user to a specific subscription topic."
        ),
        tags=["Subscription Topics"],
        responses={
            200: UserSubscriptionSerializer,
            201: UserSubscriptionSerializer,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["POST"])
    def subscribe(self, request, pk=None):
        topic = self.get_object()
        user = request.user

        existing = UserSubscription.objects.filter(
            user=user, topic=topic
        ).first()

        if existing:
            if existing.status == UserSubscription.SubscriptionStatus.ACTIVE:
                return Response(
                    {"detail": _("Already subscribed to this topic.")},
                    status=status.HTTP_400_BAD_REQUEST,
                )
            elif existing.status == UserSubscription.SubscriptionStatus.PENDING:
                return Response(
                    {"detail": _("Subscription pending confirmation.")},
                    status=status.HTTP_400_BAD_REQUEST,
                )
            else:
                existing.status = UserSubscription.SubscriptionStatus.ACTIVE
                existing.unsubscribed_at = None
                existing.save()

                response_serializer_class = self.get_response_serializer()
                response_serializer = response_serializer_class(existing)
                return Response(response_serializer.data)

        subscription_data = {
            "user": user,
            "topic": topic,
            "status": (
                UserSubscription.SubscriptionStatus.PENDING
                if topic.requires_confirmation
                else UserSubscription.SubscriptionStatus.ACTIVE
            ),
        }

        if topic.requires_confirmation:
            subscription_data["confirmation_token"] = get_random_string(64)

        subscription = UserSubscription.objects.create(**subscription_data)

        if topic.requires_confirmation:
            success = send_subscription_confirmation(subscription, user)
            if not success:
                logger.warning(
                    f"Failed to send confirmation email for subscription {subscription.id}"
                )

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(subscription)
        return Response(
            response_serializer.data, status=status.HTTP_201_CREATED
        )

    @extend_schema(
        operation_id="unsubscribeFromTopic",
        summary=_("Unsubscribe from a topic"),
        description=_(
            "Unsubscribe the current user from a specific subscription topic."
        ),
        tags=["Subscription Topics"],
        responses={
            200: None,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["POST"])
    def unsubscribe(self, request, pk=None):
        topic = self.get_object()
        user = request.user

        try:
            subscription = UserSubscription.objects.get(
                user=user,
                topic=topic,
                status=UserSubscription.SubscriptionStatus.ACTIVE,
            )
            subscription.unsubscribe()
            return Response(
                {"detail": _("Successfully unsubscribed.")},
                status=status.HTTP_200_OK,
            )
        except UserSubscription.DoesNotExist:
            return Response(
                {"detail": _("You are not subscribed to this topic.")},
                status=status.HTTP_400_BAD_REQUEST,
            )


user_subscription_req_serializers: RequestSerializersConfig = {
    "create": UserSubscriptionWriteSerializer,
    "update": UserSubscriptionWriteSerializer,
    "partial_update": UserSubscriptionWriteSerializer,
    "bulk_update": BulkSubscriptionSerializer,
    "confirm": None,
}

user_subscription_res_serializers: ResponseSerializersConfig = {
    "create": UserSubscriptionDetailSerializer,
    "list": UserSubscriptionSerializer,
    "retrieve": UserSubscriptionDetailSerializer,
    "update": UserSubscriptionDetailSerializer,
    "partial_update": UserSubscriptionDetailSerializer,
    "bulk_update": None,
    "confirm": UserSubscriptionDetailSerializer,
}


@extend_schema_view(
    **create_schema_view_config(
        model_class=UserSubscription,
        display_config={
            "tag": "User Subscription",
        },
        request_serializers=user_subscription_req_serializers,
        response_serializers=user_subscription_res_serializers,
    )
)
class UserSubscriptionViewSet(BaseModelViewSet):
    queryset = UserSubscription.objects.none()
    request_serializers = user_subscription_req_serializers
    response_serializers = user_subscription_res_serializers

    permission_classes = [IsOwnerOrAdmin]
    filterset_class = UserSubscriptionFilter
    ordering_fields = [
        "subscribed_at",
        "unsubscribed_at",
        "created_at",
        "updated_at",
        "status",
        "topic__category",
    ]
    ordering = ["-subscribed_at"]
    search_fields = [
        "topic__translations__name",
        "topic__translations__description",
        "topic__slug",
    ]

    def get_queryset(self):
        return UserSubscription.objects.filter(
            user=self.request.user
        ).select_related("topic")

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)

    @extend_schema(
        operation_id="bulkUpdateUserSubscriptions",
        summary=_("Bulk update user subscriptions"),
        description=_("Subscribe or unsubscribe from multiple topics at once."),
        tags=["User Subscriptions"],
        request=BulkSubscriptionSerializer,
        responses={
            200: None,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
        },
    )
    @action(detail=False, methods=["POST"])
    def bulk_update(self, request):
        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(data=request.data)
        request_serializer.is_valid(raise_exception=True)

        topic_ids = request_serializer.validated_data["topic_ids"]
        action = request_serializer.validated_data["action"]
        user = request.user

        existing_topics = list(
            SubscriptionTopic.objects.filter(id__in=topic_ids, is_active=True)
        )

        if len(existing_topics) != len(topic_ids):
            existing_ids = [topic.id for topic in existing_topics]
            invalid_ids = [tid for tid in topic_ids if tid not in existing_ids]
            return Response(
                {
                    "detail": _("Invalid or inactive topic IDs: {}").format(
                        invalid_ids
                    )
                },
                status=status.HTTP_400_BAD_REQUEST,
            )

        results = {"success": [], "failed": [], "already_processed": []}

        with transaction.atomic():
            for topic in existing_topics:
                try:
                    if action == "subscribe":
                        subscription, created = (
                            UserSubscription.objects.get_or_create(
                                user=user,
                                topic=topic,
                                defaults={
                                    "status": UserSubscription.SubscriptionStatus.ACTIVE
                                },
                            )
                        )

                        if created:
                            results["success"].append(topic.name)
                        elif (
                            subscription.status
                            != UserSubscription.SubscriptionStatus.ACTIVE
                        ):
                            subscription.status = (
                                UserSubscription.SubscriptionStatus.ACTIVE
                            )
                            subscription.unsubscribed_at = None
                            subscription.save()
                            results["success"].append(topic.name)
                        else:
                            results["already_processed"].append(topic.name)

                    else:
                        try:
                            subscription = UserSubscription.objects.get(
                                user=user,
                                topic=topic,
                                status=UserSubscription.SubscriptionStatus.ACTIVE,
                            )
                            subscription.unsubscribe()
                            results["success"].append(topic.name)
                        except UserSubscription.DoesNotExist:
                            results["already_processed"].append(topic.name)

                except Exception as e:
                    results["failed"].append(
                        {"topic": topic.name, "error": str(e)}
                    )

        return Response(results)

    @extend_schema(
        operation_id="confirmUserSubscription",
        summary=_("Confirm a user subscription"),
        description=_(
            "Confirm a pending subscription using the confirmation token."
        ),
        tags=["User Subscriptions"],
        responses={
            200: UserSubscriptionDetailSerializer,
            400: ErrorResponseSerializer,
            401: ErrorResponseSerializer,
            403: ErrorResponseSerializer,
            404: ErrorResponseSerializer,
        },
    )
    @action(detail=True, methods=["POST"])
    def confirm(self, request, pk=None):
        try:
            subscription = self.get_object()
            token = request.data.get("token")

            if (
                subscription.status
                != UserSubscription.SubscriptionStatus.PENDING
            ):
                return Response(
                    {"detail": _("Subscription is not pending confirmation.")},
                    status=status.HTTP_400_BAD_REQUEST,
                )

            if subscription.confirmation_token != token:
                return Response(
                    {"detail": _("Invalid confirmation token.")},
                    status=status.HTTP_400_BAD_REQUEST,
                )

            subscription.activate()

            response_serializer_class = self.get_response_serializer()
            response_serializer = response_serializer_class(
                subscription, context=self.get_serializer_context()
            )
            return Response(response_serializer.data)
        except Exception as e:
            return Response(
                {"detail": str(e)}, status=status.HTTP_400_BAD_REQUEST
            )


class UnsubscribeView(APIView):
    permission_classes = []

    class UnsubscribeSerializer(serializers.Serializer):
        message = serializers.CharField()
        topic = serializers.CharField(required=False)
        user_email = serializers.EmailField(required=False)
        topic_slug = serializers.CharField(required=False)
        error = serializers.CharField(required=False)

    serializer_class = UnsubscribeSerializer

    @extend_schema(
        operation_id="unsubscribeViaLink",
        summary=_("Unsubscribe via email link"),
        description=_(
            "Unsubscribe from a subscription topic using an email unsubscribe link."
        ),
        tags=["User Subscriptions"],
        responses={
            200: UnsubscribeSerializer,
            400: ErrorResponseSerializer,
        },
    )
    def get(self, request, uidb64: str, token: str, topic_slug: str):
        try:
            uid = urlsafe_base64_decode(uidb64).decode()
            user = User.objects.get(pk=uid)
        except (TypeError, ValueError, OverflowError, User.DoesNotExist):
            return Response(
                {"error": "Invalid unsubscribe link"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        if not default_token_generator.check_token(user, token):
            return Response(
                {"error": "Invalid or expired unsubscribe link"},
                status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            subscription = UserSubscription.objects.get(
                user=user,
                topic__slug=topic_slug,
                status=UserSubscription.SubscriptionStatus.ACTIVE,
            )
            subscription.unsubscribe()

            return Response(
                {
                    "message": "Successfully unsubscribed",
                    "topic": subscription.topic.name,
                    "user_email": user.email,
                }
            )

        except UserSubscription.DoesNotExist:
            return Response(
                {"message": "Already unsubscribed", "topic_slug": topic_slug}
            )



from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase

from product.factories.product import ProductFactory
from product.factories.review import ProductReviewFactory
from user.factories.account import UserAccountFactory

languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]
default_language = settings.PARLER_DEFAULT_LANGUAGE_CODE
User = get_user_model()


class ProductReviewModelTestCase(TestCase):
    def setUp(self):
        self.user = UserAccountFactory(num_addresses=0)
        self.product = ProductFactory(num_images=0, num_reviews=0)
        self.product_review = ProductReviewFactory(
            product=self.product,
            user=self.user,
            rate=5,
            status="New",
        )

    def test_fields(self):
        self.assertEqual(self.product_review.product, self.product)
        self.assertEqual(self.product_review.user, self.user)
        self.assertEqual(self.product_review.rate, 5)
        self.assertEqual(self.product_review.status, "New")

    def test_str_representation(self):
        comment_snippet = (
            (
                self.product_review.safe_translation_getter(
                    "comment", any_language=True
                )[:50]
                + "..."
            )
            if self.product_review.comment
            else "No Comment"
        )
        self.assertEqual(
            str(self.product_review),
            f"Review by {self.user.email} on {self.product}: {comment_snippet}",
        )



from datetime import timedelta
from unittest.mock import patch

import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.db import connection
from django.utils import timezone
from djmoney.money import Money

from cart.factories import CartFactory, CartItemFactory
from cart.models import Cart, CartItem
from country.factories import CountryFactory
from product.factories import ProductFactory
from user.factories import UserAccountFactory

User = get_user_model()


@pytest.fixture
def country():
    return CountryFactory(alpha_2="US", name="United States")


@pytest.fixture
def user(country):
    return UserAccountFactory(country=country)


@pytest.fixture
def cart(user):
    return CartFactory(user=user)


@pytest.fixture
def guest_cart():
    return CartFactory(user=None, session_key="guest123")


@pytest.fixture
def product():
    return ProductFactory(price=Money(50, settings.DEFAULT_CURRENCY))


@pytest.fixture
def expensive_product():
    return ProductFactory(price=Money(150, settings.DEFAULT_CURRENCY))


@pytest.fixture
def discounted_product():
    return ProductFactory(
        price=Money(30, settings.DEFAULT_CURRENCY), discount_percent=20
    )


@pytest.mark.django_db
class TestCartItemQuerySet:
    def test_for_cart(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)
        other_cart = CartFactory()
        CartItemFactory(cart=other_cart, product=product)

        items_for_cart = CartItem.objects.for_cart(cart)
        assert item in items_for_cart
        assert items_for_cart.count() == 1

    def test_for_product(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)
        other_product = ProductFactory()
        CartItemFactory(cart=cart, product=other_product)

        items_for_product = CartItem.objects.for_product(product)
        assert item in items_for_product
        assert items_for_product.count() == 1

    def test_for_authenticated_user(self, user, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        items_for_user = CartItem.objects.for_user(user)
        assert item in items_for_user

    def test_for_anonymous_user(self):
        anonymous_user = AnonymousUser()

        items_for_user = CartItem.objects.for_user(anonymous_user)
        assert items_for_user.count() == 0

    def test_with_product_data(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        items_with_data = CartItem.objects.with_product_data()
        item_with_data = items_with_data.get(id=item.id)

        assert hasattr(item_with_data, "product")
        assert hasattr(item_with_data, "cart")

    def test_total_quantity(self, cart, product):
        CartItemFactory(cart=cart, product=product, quantity=3)
        CartItemFactory(cart=cart, product=ProductFactory(), quantity=5)

        total = CartItem.objects.total_quantity()
        assert total == 8

    def test_total_quantity_empty(self):
        total = CartItem.objects.total_quantity()
        assert total == 0

    def test_by_product_popularity(self, cart, product):
        _ = CartItemFactory(cart=cart, product=product, quantity=5)
        cart2 = CartFactory(user=None, session_key="guest456")
        _ = CartItemFactory(cart=cart2, product=product, quantity=3)

        other_product = ProductFactory()
        CartItemFactory(cart=cart, product=other_product, quantity=2)

        popularity_data = CartItem.objects.by_product_popularity()

        first_item = popularity_data.first()
        assert first_item["product"] == product.id
        assert first_item["total_quantity"] == 8
        assert first_item["cart_count"] == 2

    def test_high_quantity_default_threshold(self, cart, product):
        high_quantity_item = CartItemFactory(
            cart=cart, product=product, quantity=6
        )
        low_quantity_item = CartItemFactory(
            cart=cart, product=ProductFactory(), quantity=3
        )

        high_quantity_items = CartItem.objects.high_quantity()
        assert high_quantity_item in high_quantity_items
        assert low_quantity_item not in high_quantity_items

    def test_high_quantity_custom_threshold(self, cart, product):
        item = CartItemFactory(cart=cart, product=product, quantity=8)

        high_quantity_items = CartItem.objects.high_quantity(threshold=10)
        assert item not in high_quantity_items

        high_quantity_items = CartItem.objects.high_quantity(threshold=7)
        assert item in high_quantity_items

    def test_recent_items_default_7_days(self, cart, product):
        recent_item = CartItemFactory(cart=cart, product=product)
        recent_item.created_at = timezone.now() - timedelta(days=3)
        recent_item.save()

        old_item = CartItemFactory(cart=cart, product=ProductFactory())
        old_item.created_at = timezone.now() - timedelta(days=10)
        old_item.save()

        recent_items = CartItem.objects.recent()
        assert recent_item in recent_items
        assert old_item not in recent_items

    def test_recent_items_custom_days(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)
        item.created_at = timezone.now() - timedelta(days=5)
        item.save()

        recent_items = CartItem.objects.recent(days=3)
        assert item not in recent_items

        recent_items = CartItem.objects.recent(days=7)
        assert item in recent_items

    def test_by_date_range(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)
        item.created_at = timezone.now() - timedelta(days=5)
        item.save()

        start_date = (timezone.now() - timedelta(days=7)).date()
        end_date = (timezone.now() - timedelta(days=3)).date()

        items_in_range = CartItem.objects.by_date_range(start_date, end_date)
        assert item in items_in_range

    def test_by_quantity_range(self, cart, product):
        low_item = CartItemFactory(cart=cart, product=product, quantity=2)
        mid_item = CartItemFactory(
            cart=cart, product=ProductFactory(), quantity=5
        )
        high_item = CartItemFactory(
            cart=cart, product=ProductFactory(), quantity=8
        )

        items_in_range = CartItem.objects.by_quantity_range(3, 6)
        assert low_item not in items_in_range
        assert mid_item in items_in_range
        assert high_item not in items_in_range

    def test_by_price_range(self, cart):
        cheap_product = ProductFactory(
            price=Money(20, settings.DEFAULT_CURRENCY)
        )
        mid_product = ProductFactory(price=Money(75, settings.DEFAULT_CURRENCY))
        expensive_product = ProductFactory(
            price=Money(200, settings.DEFAULT_CURRENCY)
        )

        cheap_item = CartItemFactory(cart=cart, product=cheap_product)
        mid_item = CartItemFactory(cart=cart, product=mid_product)
        expensive_item = CartItemFactory(cart=cart, product=expensive_product)

        items_in_range = CartItem.objects.by_price_range(50, 100)
        assert cheap_item not in items_in_range
        assert mid_item in items_in_range
        assert expensive_item not in items_in_range

    def test_expensive_items_default_threshold(
        self, cart, product, expensive_product
    ):
        regular_item = CartItemFactory(cart=cart, product=product)
        expensive_item = CartItemFactory(cart=cart, product=expensive_product)

        expensive_items = CartItem.objects.expensive_items()
        assert regular_item not in expensive_items
        assert expensive_item in expensive_items

    def test_expensive_items_custom_threshold(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        expensive_items = CartItem.objects.expensive_items(threshold=30)
        assert item in expensive_items

        expensive_items = CartItem.objects.expensive_items(threshold=80)
        assert item not in expensive_items

    @patch("extra_settings.models.Setting.get")
    def test_in_active_carts(self, mock_setting, cart, product):
        mock_setting.return_value = 24

        active_item = CartItemFactory(cart=cart, product=product)

        abandoned_cart = CartFactory(user=None, session_key="guest_abandoned")
        abandoned_item = CartItemFactory(
            cart=abandoned_cart, product=ProductFactory()
        )

        active_time = timezone.now() - timedelta(hours=1)
        abandoned_time = timezone.now() - timedelta(hours=25)

        with connection.cursor() as cursor:
            cursor.execute(
                "UPDATE cart_cart SET last_activity = %s WHERE id = %s",
                [active_time, cart.id],
            )
            cursor.execute(
                "UPDATE cart_cart SET last_activity = %s WHERE id = %s",
                [abandoned_time, abandoned_cart.id],
            )

        cart.refresh_from_db()
        abandoned_cart.refresh_from_db()

        active_items = CartItem.objects.in_active_carts()

        assert active_item in active_items
        assert abandoned_item not in active_items

    @patch("extra_settings.models.Setting.get")
    def test_in_abandoned_carts(self, mock_setting, cart, product):
        mock_setting.return_value = 24

        abandoned_item = CartItemFactory(cart=cart, product=product)

        active_cart = CartFactory(user=None, session_key="guest_active")
        active_item = CartItemFactory(
            cart=active_cart, product=ProductFactory()
        )

        abandoned_time = timezone.now() - timedelta(hours=25)
        active_time = timezone.now() - timedelta(hours=1)

        with connection.cursor() as cursor:
            cursor.execute(
                "UPDATE cart_cart SET last_activity = %s WHERE id = %s",
                [abandoned_time, cart.id],
            )
            cursor.execute(
                "UPDATE cart_cart SET last_activity = %s WHERE id = %s",
                [active_time, active_cart.id],
            )

        cart.refresh_from_db()
        active_cart.refresh_from_db()

        abandoned_items = CartItem.objects.in_abandoned_carts()
        assert abandoned_item in abandoned_items
        assert active_item not in abandoned_items

    def test_with_discounts(self, cart):
        regular_product = ProductFactory(discount_percent=0)
        discounted_product = ProductFactory(discount_percent=20)

        regular_item = CartItemFactory(cart=cart, product=regular_product)
        discounted_item = CartItemFactory(cart=cart, product=discounted_product)

        items_with_discounts = CartItem.objects.with_discounts()
        assert regular_item not in items_with_discounts
        assert discounted_item in items_with_discounts


@pytest.mark.django_db
class TestCartItemManager:
    def test_manager_delegates_to_queryset_for_cart(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        items_for_cart = CartItem.objects.for_cart(cart)
        assert item in items_for_cart

    def test_manager_delegates_to_queryset_for_product(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        items_for_product = CartItem.objects.for_product(product)
        assert item in items_for_product

    def test_manager_delegates_to_queryset_for_user(self, user, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        items_for_user = CartItem.objects.for_user(user)
        assert item in items_for_user

    def test_manager_delegates_to_queryset_with_product_data(
        self, cart, product
    ):
        _ = CartItemFactory(cart=cart, product=product)

        items_with_data = CartItem.objects.with_product_data()
        assert items_with_data.count() == 1

    def test_manager_delegates_to_queryset_total_quantity(self, cart, product):
        CartItemFactory(cart=cart, product=product, quantity=3)

        total = CartItem.objects.total_quantity()
        assert total == 3

    def test_manager_delegates_to_queryset_by_product_popularity(
        self, cart, product
    ):
        CartItemFactory(cart=cart, product=product, quantity=5)

        popularity_data = CartItem.objects.by_product_popularity()
        assert popularity_data.count() == 1

    def test_manager_delegates_to_queryset_high_quantity(self, cart, product):
        item = CartItemFactory(cart=cart, product=product, quantity=6)

        high_quantity_items = CartItem.objects.high_quantity()
        assert item in high_quantity_items

    def test_manager_delegates_to_queryset_recent(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        recent_items = CartItem.objects.recent()
        assert item in recent_items

    def test_manager_delegates_to_queryset_by_date_range(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)
        item.created_at = timezone.now() - timedelta(days=5)
        item.save()

        start_date = (timezone.now() - timedelta(days=7)).date()
        end_date = (timezone.now() - timedelta(days=3)).date()

        items_in_range = CartItem.objects.by_date_range(start_date, end_date)
        assert item in items_in_range

    def test_manager_delegates_to_queryset_by_quantity_range(
        self, cart, product
    ):
        item = CartItemFactory(cart=cart, product=product, quantity=5)

        items_in_range = CartItem.objects.by_quantity_range(3, 6)
        assert item in items_in_range

    def test_manager_delegates_to_queryset_by_price_range(self, cart, product):
        item = CartItemFactory(cart=cart, product=product)

        items_in_range = CartItem.objects.by_price_range(40, 60)
        assert item in items_in_range

    def test_manager_delegates_to_queryset_expensive_items(
        self, cart, expensive_product
    ):
        item = CartItemFactory(cart=cart, product=expensive_product)

        expensive_items = CartItem.objects.expensive_items()
        assert item in expensive_items

    def test_manager_delegates_to_queryset_in_active_carts(self, cart, product):
        Cart.objects.filter(id=cart.id).update(last_activity=timezone.now())
        item = CartItemFactory(cart=cart, product=product)

        active_items = CartItem.objects.in_active_carts()
        assert item in active_items

    def test_manager_delegates_to_queryset_in_abandoned_carts(
        self, cart, product
    ):
        Cart.objects.filter(id=cart.id).update(
            last_activity=timezone.now() - timedelta(hours=25)
        )
        item = CartItemFactory(cart=cart, product=product)

        abandoned_items = CartItem.objects.in_abandoned_carts()
        assert item in abandoned_items

    def test_manager_delegates_to_queryset_with_discounts(
        self, cart, discounted_product
    ):
        item = CartItemFactory(cart=cart, product=discounted_product)

        items_with_discounts = CartItem.objects.with_discounts()
        assert item in items_with_discounts


@pytest.mark.django_db
class TestCartItemQuerySetEdgeCases:
    def test_by_product_popularity_with_no_items(self):
        popularity_data = CartItem.objects.by_product_popularity()
        assert popularity_data.count() == 0

    def test_total_quantity_aggregation_accuracy(self, cart):
        CartItemFactory(cart=cart, product=ProductFactory(), quantity=3)
        CartItemFactory(cart=cart, product=ProductFactory(), quantity=7)
        CartItemFactory(cart=cart, product=ProductFactory(), quantity=2)

        total = CartItem.objects.total_quantity()
        assert total == 12

    def test_price_range_boundary_conditions(self, cart):
        product_at_min = ProductFactory(
            price=Money(50, settings.DEFAULT_CURRENCY)
        )
        product_at_max = ProductFactory(
            price=Money(100, settings.DEFAULT_CURRENCY)
        )
        product_below = ProductFactory(
            price=Money(49.99, settings.DEFAULT_CURRENCY)
        )
        product_above = ProductFactory(
            price=Money(100.01, settings.DEFAULT_CURRENCY)
        )

        item_at_min = CartItemFactory(cart=cart, product=product_at_min)
        item_at_max = CartItemFactory(cart=cart, product=product_at_max)
        item_below = CartItemFactory(cart=cart, product=product_below)
        item_above = CartItemFactory(cart=cart, product=product_above)

        items_in_range = CartItem.objects.by_price_range(50, 100)
        assert item_at_min in items_in_range
        assert item_at_max in items_in_range
        assert item_below not in items_in_range
        assert item_above not in items_in_range

    def test_quantity_range_boundary_conditions(self, cart, product):
        item_at_min = CartItemFactory(cart=cart, product=product, quantity=5)
        item_at_max = CartItemFactory(
            cart=cart, product=ProductFactory(), quantity=10
        )
        item_below = CartItemFactory(
            cart=cart, product=ProductFactory(), quantity=4
        )
        item_above = CartItemFactory(
            cart=cart, product=ProductFactory(), quantity=11
        )

        items_in_range = CartItem.objects.by_quantity_range(5, 10)
        assert item_at_min in items_in_range
        assert item_at_max in items_in_range
        assert item_below not in items_in_range
        assert item_above not in items_in_range

    def test_with_discounts_zero_discount_excluded(self, cart):
        zero_discount_product = ProductFactory(discount_percent=0)
        item = CartItemFactory(cart=cart, product=zero_discount_product)

        items_with_discounts = CartItem.objects.with_discounts()
        assert item not in items_with_discounts



from decimal import Decimal
from unittest import TestCase, mock

from django.conf import settings
from djmoney.money import Money
from djstripe.models import PaymentIntent

from order.enum.status import PaymentStatus
from order.payment import (
    PayPalPaymentProvider,
    StripePaymentProvider,
    get_payment_provider,
)


class PaymentModuleTestCase(TestCase):
    def test_payment_status_enum(self):
        self.assertEqual(PaymentStatus.PENDING.value, "PENDING")
        self.assertEqual(PaymentStatus.PROCESSING.value, "PROCESSING")
        self.assertEqual(PaymentStatus.COMPLETED.value, "COMPLETED")
        self.assertEqual(PaymentStatus.FAILED.value, "FAILED")
        self.assertEqual(PaymentStatus.REFUNDED.value, "REFUNDED")
        self.assertEqual(
            PaymentStatus.PARTIALLY_REFUNDED.value, "PARTIALLY_REFUNDED"
        )
        self.assertEqual(PaymentStatus.CANCELED.value, "CANCELED")

    @mock.patch("order.payment.settings")
    def test_stripe_payment_provider_init(self, mock_settings):
        mock_settings.STRIPE_TEST_SECRET_KEY = "test_api_key"
        mock_settings.DJSTRIPE_WEBHOOK_SECRET = "test_webhook_secret"
        mock_settings.STRIPE_LIVE_MODE = False

        provider = StripePaymentProvider()

        self.assertEqual(provider.api_key, "test_api_key")
        self.assertEqual(provider.webhook_secret, "test_webhook_secret")

    @mock.patch("order.payment.settings")
    def test_paypal_payment_provider_init(self, mock_settings):
        mock_settings.PAYPAL_CLIENT_ID = "test_client_id"
        mock_settings.PAYPAL_CLIENT_SECRET = "test_client_secret"

        provider = PayPalPaymentProvider()

        self.assertEqual(provider.client_id, "test_client_id")
        self.assertEqual(provider.client_secret, "test_client_secret")

    @mock.patch("order.payment.stripe.PaymentIntent.create")
    @mock.patch("order.payment.logger")
    def test_stripe_process_payment(self, mock_logger, mock_stripe_create):
        mock_payment_intent = mock.Mock()
        mock_payment_intent.id = "pi_test_order_id_mock"
        mock_payment_intent.status = "succeeded"
        mock_payment_intent.client_secret = "pi_test_secret"
        mock_payment_intent.next_action = None
        mock_stripe_create.return_value = mock_payment_intent

        with mock.patch("order.payment.PaymentIntent.sync_from_stripe_data"):
            provider = StripePaymentProvider()
            amount = Money(
                amount=Decimal("100.00"), currency=settings.DEFAULT_CURRENCY
            )
            order_id = "test_order_id"

            success, payment_data = provider.process_payment(amount, order_id)

            self.assertTrue(success)
            self.assertEqual(
                payment_data["payment_id"], "pi_test_order_id_mock"
            )
            self.assertEqual(payment_data["status"], PaymentStatus.COMPLETED)

    @mock.patch("order.payment.logger")
    def test_paypal_process_payment(self, mock_logger):
        provider = PayPalPaymentProvider()
        amount = Money(
            amount=Decimal("100.00"), currency=settings.DEFAULT_CURRENCY
        )
        order_id = "test_order_id"

        success, payment_data = provider.process_payment(amount, order_id)

        self.assertTrue(success)
        self.assertEqual(payment_data["payment_id"], f"PP_{order_id}_mock")
        self.assertEqual(payment_data["status"], PaymentStatus.COMPLETED)
        self.assertEqual(payment_data["amount"], str(amount.amount))
        self.assertEqual(payment_data["currency"], str(amount.currency))
        self.assertEqual(payment_data["provider"], "paypal")

        mock_logger.info.assert_called_once()

    @mock.patch("order.payment.stripe.Refund.create")
    @mock.patch("order.payment.logger")
    def test_stripe_refund_payment(self, mock_logger, mock_stripe_refund):
        mock_refund = mock.Mock()
        mock_refund.id = "re_test_payment_id_mock"
        mock_refund.status = "succeeded"
        mock_stripe_refund.return_value = mock_refund

        with mock.patch("order.payment.Refund.sync_from_stripe_data"):
            provider = StripePaymentProvider()
            payment_id = "test_payment_id"
            amount = Money(
                amount=Decimal("50.00"), currency=settings.DEFAULT_CURRENCY
            )

            success, refund_data = provider.refund_payment(payment_id, amount)

            self.assertTrue(success)
            self.assertEqual(
                refund_data["refund_id"], "re_test_payment_id_mock"
            )

    @mock.patch("order.payment.stripe.PaymentIntent.retrieve")
    @mock.patch("order.payment.PaymentIntent.sync_from_stripe_data")
    @mock.patch("order.payment.logger")
    def test_stripe_get_payment_status(
        self, mock_logger, mock_sync, mock_retrieve
    ):
        mock_payment_intent = mock.Mock()
        mock_payment_intent.id = "test_payment_id"
        mock_payment_intent.status = "succeeded"
        mock_payment_intent.amount = 10000
        mock_payment_intent.currency = "eur"
        mock_payment_intent.created = 1234567890

        mock_retrieve.return_value = mock_payment_intent
        mock_djstripe_pi = mock.Mock()
        mock_djstripe_pi.created = "2024-01-01"
        mock_sync.return_value = mock_djstripe_pi

        with mock.patch(
            "order.payment.PaymentIntent.objects.get",
            side_effect=PaymentIntent.DoesNotExist,
        ):
            provider = StripePaymentProvider()
            payment_id = "test_payment_id"

            status, status_data = provider.get_payment_status(payment_id)

            self.assertEqual(status, PaymentStatus.COMPLETED)
            self.assertEqual(status_data["payment_id"], payment_id)

    def test_get_payment_provider(self):
        provider = get_payment_provider("stripe")
        self.assertIsInstance(provider, StripePaymentProvider)

        provider = get_payment_provider("paypal")
        self.assertIsInstance(provider, PayPalPaymentProvider)

        with self.assertRaises(ValueError):
            get_payment_provider("invalid_provider")



import importlib

import factory
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import Count

from product.enum.review import RateEnum, ReviewStatus
from product.models.review import ProductReview

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]

User = get_user_model()


def get_or_create_user():
    if User.objects.exists():
        user = (
            User.objects.annotate(num_reviews=Count("product_reviews"))
            .order_by("num_reviews")
            .first()
        )
    else:
        user_factory_module = importlib.import_module("user.factories.account")
        user_factory_class = user_factory_module.UserAccountFactory
        user = user_factory_class.create()
    return user


def get_or_create_product():
    if apps.get_model("product", "Product").objects.exists():
        return (
            apps.get_model("product", "Product")
            .objects.annotate(num_reviews=Count("reviews"))
            .order_by("num_reviews")
            .first()
        )
    else:
        product_factory_module = importlib.import_module(
            "product.factories.product"
        )
        product_factory_class = product_factory_module.ProductFactory
        return product_factory_class.create()


class ProductReviewTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    comment = factory.Faker("paragraph")
    master = factory.SubFactory("product.factories.review.ProductReviewFactory")

    class Meta:
        model = apps.get_model("product", "ProductReviewTranslation")
        django_get_or_create = ("language_code", "master")


class ProductReviewFactory(factory.django.DjangoModelFactory):
    product = factory.LazyFunction(get_or_create_product)
    user = factory.LazyFunction(get_or_create_user)
    rate = factory.Iterator([choice.value for choice in RateEnum])
    status = factory.Iterator([choice.value for choice in ReviewStatus])

    class Meta:
        model = ProductReview
        django_get_or_create = ("product", "user")
        skip_postgeneration_save = True

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        translations = extracted or [
            ProductReviewTranslationFactory(language_code=lang, master=self)
            for lang in available_languages
        ]

        for translation in translations:
            translation.master = self
            translation.save()



import factory
from django.apps import apps
from django.conf import settings
from faker import Faker

from core.factories import (
    CustomDjangoModelFactory,
    custom_seeding,
    SeedingResult,
)
from country.models import Country

fake = Faker()

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]

LANGUAGE_COUNTRY_MAPPING = getattr(settings, "LANGUAGE_COUNTRY_MAPPING", {})


class CountryTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    name = factory.LazyAttribute(
        lambda obj: f"Country Name ({obj.language_code})"
    )
    master = factory.SubFactory("country.factories.CountryFactory")

    class Meta:
        model = apps.get_model("country", "CountryTranslation")
        django_get_or_create = ("language_code", "master")


@custom_seeding(
    description="Creates countries based on PARLER_LANGUAGES settings (el→GR, en→GB, de→DE)",
    settings_key="PARLER_LANGUAGES",
)
class CountryFactory(CustomDjangoModelFactory):
    auto_translations = False

    unique_model_fields = [
        ("alpha_2", lambda: fake.country_code(representation="alpha-2")),
        ("alpha_3", lambda: fake.country_code(representation="alpha-3")),
        ("phone_code", lambda: fake.random_int(min=1, max=9999)),
    ]

    image_flag = factory.django.ImageField(
        filename=factory.LazyAttribute(lambda _: f"{fake.word()}.png"),
        color=factory.Faker("color"),
        width=640,
        height=480,
    )

    class Meta:
        model = Country
        django_get_or_create = ("alpha_2",)
        skip_postgeneration_save = True

    @classmethod
    def custom_seed(cls, **kwargs) -> SeedingResult:
        result = SeedingResult(created_count=0)

        verbose = kwargs.get("verbose", True)

        for lang_code in available_languages:
            if lang_code in LANGUAGE_COUNTRY_MAPPING:
                country_data = LANGUAGE_COUNTRY_MAPPING[lang_code]

                existing_country = Country.objects.filter(
                    alpha_2=country_data["alpha_2"]
                ).first()

                if existing_country:
                    if verbose:
                        print(
                            f"Country {country_data['alpha_2']} already exists, skipping..."
                        )
                    result.skipped_count += 1
                    continue

                try:
                    country = Country.objects.create(
                        alpha_2=country_data["alpha_2"],
                        alpha_3=country_data["alpha_3"],
                        iso_cc=country_data.get("iso_cc"),
                        phone_code=country_data["phone_code"],
                    )

                    for trans_lang in available_languages:
                        country_name = country_data["names"].get(
                            trans_lang, country_data["names"]["en"]
                        )
                        translation_model = apps.get_model(
                            "country", "CountryTranslation"
                        )
                        translation_model.objects.get_or_create(
                            language_code=trans_lang,
                            master=country,
                            defaults={"name": country_name},
                        )

                    result.created_count += 1
                    if verbose:
                        print(f"Created country: {country_data['alpha_2']}")

                except Exception as e:
                    error_msg = f"Failed to create country {country_data['alpha_2']}: {str(e)}"
                    result.errors.append(error_msg)
                    if verbose:
                        print(f"Error: {error_msg}")

        return result

    @factory.post_generation
    def num_regions(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            from region.factories import RegionFactory

            RegionFactory.create_batch(extracted, country=self)

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        if not hasattr(self, "_translations_created"):
            translations = extracted or [
                CountryTranslationFactory(language_code=lang, master=self)
                for lang in available_languages
            ]

            for translation in translations:
                translation.save()



import asyncio
from unittest.mock import AsyncMock, patch
from asgiref.sync import sync_to_async
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings

from blog.models import BlogComment, BlogPost
from blog.signals import notify_comment_liked, notify_comment_liked_receiver
from notification.enum import NotificationKindEnum
from notification.models.notification import Notification
from notification.models.user import NotificationUser

User = get_user_model()


class TestNotifyCommentLiked(TestCase):
    def setUp(self):
        self.blog_post = BlogPost.objects.create(slug="test-post")

        self.comment_user = User.objects.create_user(
            email="commenter@example.com",
            username="commenter",
            password="testpass123",
        )
        self.liker_user = User.objects.create_user(
            email="liker@example.com", username="liker", password="testpass123"
        )

        self.comment = BlogComment.objects.create(
            post=self.blog_post, user=self.comment_user, approved=True
        )

    def _setup_async_mocks(
        self,
        mock_sync_to_async,
        mock_user_aget,
        mock_notification_acreate,
        mock_notification_user_acreate,
        user_return_value,
    ):
        def sync_to_async_side_effect(fn):
            async def wrapped_async(*args, **kwargs):
                if asyncio.iscoroutinefunction(fn):
                    return await fn(*args, **kwargs)
                return fn(*args, **kwargs)

            return wrapped_async

        mock_sync_to_async.side_effect = sync_to_async_side_effect

        async def async_aget_mock(*args, **kwargs):
            if callable(user_return_value):
                return user_return_value(*args, **kwargs)
            return user_return_value

        mock_user_aget.side_effect = async_aget_mock

        mock_notification = AsyncMock()
        mock_notification.title = ""
        mock_notification.message = ""
        mock_notification.asave = AsyncMock()

        async def async_notification_acreate(*args, **kwargs):
            return mock_notification

        mock_notification_acreate.side_effect = async_notification_acreate

        async def async_user_acreate_mock(*args, **kwargs):
            return AsyncMock()

        mock_notification_user_acreate.side_effect = async_user_acreate_mock

        return mock_notification

    async def test_notify_comment_liked_basic(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):
            mock_notification = self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            mock_notification_acreate.assert_called_once_with(
                kind=NotificationKindEnum.INFO,
                link=f"{settings.NUXT_BASE_URL}/blog/post/{self.blog_post.id}/{self.blog_post.slug}#blog-post-comments",
            )

            mock_notification_user_acreate.assert_called_once_with(
                user=self.comment_user, notification=mock_notification
            )

    async def test_notify_comment_liked_multiple_languages(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
            patch("blog.signals.languages", ["en", "el"]),
        ):
            mock_notification = self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            self.assertEqual(
                mock_notification.set_current_language.call_count, 2
            )
            self.assertEqual(mock_notification.asave.call_count, 2)

    async def test_notify_comment_liked_english_content(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
            patch("blog.signals.languages", ["en"]),
        ):
            mock_notification = self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            expected_url = f"{settings.NUXT_BASE_URL}/blog/post/{self.blog_post.id}/{self.blog_post.slug}#blog-post-comments"
            expected_title = f"<a href='{expected_url}'>Comment</a> Liked!"
            expected_message = f"Your comment was liked by {self.liker_user.username or self.liker_user.email}."

            self.assertEqual(mock_notification.title, expected_title)
            self.assertEqual(mock_notification.message, expected_message)

    async def test_notify_comment_liked_greek_content(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
            patch("blog.signals.languages", ["el"]),
        ):
            mock_notification = self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            expected_url = f"{settings.NUXT_BASE_URL}/blog/post/{self.blog_post.id}/{self.blog_post.slug}#blog-post-comments"
            expected_title = (
                f"Το <a href='{expected_url}'>σχόλιο</a> σου πήρε like!"
            )
            expected_message = f"Το σχόλιο σου άρεσε στον χρήστη {self.liker_user.username or self.liker_user.email}."

            self.assertEqual(mock_notification.title, expected_title)
            self.assertEqual(mock_notification.message, expected_message)

    async def test_notify_comment_liked_user_without_username(self):
        with patch(
            "user.managers.account.UserNameGenerator"
        ) as mock_username_generator:
            mock_username_generator.return_value.generate_username.return_value = ""
            liker_no_username = await sync_to_async(User.objects.create_user)(
                email="no_username@example.com", password="testpass123"
            )

        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
            patch("blog.signals.languages", ["en"]),
        ):
            mock_notification = self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                liker_no_username,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={liker_no_username.id},
            )

            expected_message = (
                f"Your comment was liked by {liker_no_username.email}."
            )

            self.assertEqual(mock_notification.message, expected_message)

    async def test_notify_comment_liked_multiple_users(self):
        liker2 = await sync_to_async(User.objects.create_user)(
            email="liker2@example.com",
            username="liker2",
            password="testpass123",
        )

        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):

            def aget_side_effect(id):
                if id == self.liker_user.id:
                    return self.liker_user
                elif id == liker2.id:
                    return liker2

            self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                aget_side_effect,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id, liker2.id},
            )

            self.assertEqual(mock_notification_acreate.call_count, 2)
            self.assertEqual(mock_notification_user_acreate.call_count, 2)

    async def test_notify_comment_liked_self_like(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):
            self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.comment_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.comment_user.id},
            )

            mock_notification_acreate.assert_not_called()
            mock_notification_user_acreate.assert_not_called()

    def test_receiver_wrong_action(self):
        with patch("blog.signals.async_to_sync") as mock_async_to_sync:
            notify_comment_liked_receiver(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_remove",
                reverse=False,
                pk_set={self.liker_user.id},
            )
            mock_async_to_sync.assert_not_called()

    def test_receiver_reverse_action(self):
        with patch("blog.signals.async_to_sync") as mock_async_to_sync:
            notify_comment_liked_receiver(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=True,
                pk_set={self.liker_user.id},
            )
            mock_async_to_sync.assert_not_called()

    async def test_notify_comment_liked_empty_pk_set(self):
        with patch.object(
            Notification.objects, "acreate"
        ) as mock_notification_acreate:
            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set=set(),
            )
            mock_notification_acreate.assert_not_called()

    async def test_notify_comment_liked_no_comment_user(self):
        comment_no_user = await sync_to_async(BlogComment.objects.create)(
            post=self.blog_post, user=None, approved=True
        )

        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):
            self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=comment_no_user,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            mock_notification_acreate.assert_not_called()
            mock_notification_user_acreate.assert_not_called()

    async def test_notify_comment_liked_url_generation(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):
            self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            expected_url = f"{settings.NUXT_BASE_URL}/blog/post/{self.blog_post.id}/{self.blog_post.slug}#blog-post-comments"

            mock_notification_acreate.assert_called_once_with(
                kind=NotificationKindEnum.INFO, link=expected_url
            )

    @override_settings(NUXT_BASE_URL="https://test.example.com")
    async def test_notify_comment_liked_custom_base_url(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):
            self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            expected_url = f"https://test.example.com/blog/post/{self.blog_post.id}/{self.blog_post.slug}#blog-post-comments"

            mock_notification_acreate.assert_called_once_with(
                kind=NotificationKindEnum.INFO, link=expected_url
            )

    def test_signal_registration(self):
        from django.db.models.signals import m2m_changed

        all_receivers = [
            r
            for sublist in m2m_changed._live_receivers(
                sender=BlogComment.likes.through
            )
            for r in sublist
        ]

        self.assertGreater(len(all_receivers), 0)

        handler_found = any(
            hasattr(r, "__name__")
            and r.__name__ == "notify_comment_liked_receiver"
            for r in all_receivers
        )
        self.assertTrue(
            handler_found,
            "notify_comment_liked_receiver signal handler not found",
        )

    def test_languages_configuration(self):
        from blog.signals import languages

        self.assertIsInstance(languages, list)
        for lang in languages:
            self.assertIsInstance(lang, str)
            self.assertGreater(len(lang), 0)

    async def test_notify_comment_liked_error_handling(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(Notification.objects, "acreate"),
        ):

            def sync_to_async_side_effect(fn):
                async def wrapped_async(*args, **kwargs):
                    return fn(*args, **kwargs)

                return wrapped_async

            mock_sync_to_async.side_effect = sync_to_async_side_effect

            mock_user_aget.side_effect = User.DoesNotExist("Database error")

            try:
                await notify_comment_liked(
                    sender=BlogComment.likes.through,
                    instance=self.comment,
                    action="post_add",
                    reverse=False,
                    pk_set={self.liker_user.id},
                )
            except Exception as e:
                self.fail(
                    f"Signal handler should handle errors gracefully, but raised: {e}"
                )

    async def test_sync_to_async_usage(self):
        with (
            patch("blog.signals.sync_to_async") as mock_sync_to_async,
            patch.object(User.objects, "aget") as mock_user_aget,
            patch.object(
                Notification.objects, "acreate"
            ) as mock_notification_acreate,
            patch.object(
                NotificationUser.objects, "acreate"
            ) as mock_notification_user_acreate,
        ):
            self._setup_async_mocks(
                mock_sync_to_async,
                mock_user_aget,
                mock_notification_acreate,
                mock_notification_user_acreate,
                self.liker_user,
            )

            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=self.comment,
                action="post_add",
                reverse=False,
                pk_set={self.liker_user.id},
            )

            self.assertGreater(mock_sync_to_async.call_count, 0)
            mock_user_aget.assert_called()
            mock_notification_acreate.assert_called()
            mock_notification_user_acreate.assert_called()


class TestBlogSignalsIntegration(TestCase):
    def setUp(self):
        self.blog_post = BlogPost.objects.create()
        self.user1 = User.objects.create_user(
            email="user1@example.com", username="user1", password="testpass123"
        )
        self.user2 = User.objects.create_user(
            email="user2@example.com", username="user2", password="testpass123"
        )
        self.comment = BlogComment.objects.create(
            post=self.blog_post, user=self.user1, approved=True
        )

    def test_signal_sender_configuration(self):
        self.assertEqual(
            BlogComment.likes.through._meta.label, "blog.BlogComment_likes"
        )

    def test_m2m_field_configuration(self):
        likes_field = BlogComment._meta.get_field("likes")
        self.assertEqual(likes_field.related_model, User)
        self.assertTrue(hasattr(likes_field.remote_field, "through"))

    async def test_notification_enum_usage(self):
        self.assertEqual(NotificationKindEnum.INFO.value, "INFO")
        valid_choices = [choice[0] for choice in NotificationKindEnum.choices]
        self.assertIn(NotificationKindEnum.INFO.value, valid_choices)


class TestBlogSignalsEdgeCases(TestCase):
    def setUp(self):
        self.blog_post = BlogPost.objects.create()
        self.user = User.objects.create_user(
            email="test@example.com",
            username="testuser",
            password="testpass123",
        )

    async def test_notify_with_none_pk_set(self):
        comment = await sync_to_async(BlogComment.objects.create)(
            post=self.blog_post, user=self.user, approved=True
        )

        with patch.object(
            Notification.objects, "acreate"
        ) as mock_notification_acreate:
            await notify_comment_liked(
                sender=BlogComment.likes.through,
                instance=comment,
                action="post_add",
                reverse=False,
                pk_set=None,
            )
            mock_notification_acreate.assert_not_called()

    async def test_notify_with_pre_add_action(self):
        comment = await sync_to_async(BlogComment.objects.create)(
            post=self.blog_post, user=self.user, approved=True
        )

        with patch("blog.signals.async_to_sync") as mock_async_to_sync:
            notify_comment_liked_receiver(
                sender=BlogComment.likes.through,
                instance=comment,
                action="pre_add",
                reverse=False,
                pk_set={self.user.id},
            )
            mock_async_to_sync.assert_not_called()

    async def test_async_error_isolation(self):
        comment = await sync_to_async(BlogComment.objects.create)(
            post=self.blog_post, user=self.user, approved=True
        )

        with patch("blog.signals.sync_to_async") as mock_sync_to_async:
            mock_sync_to_async.side_effect = Exception(
                "Async conversion failed"
            )

            try:
                await notify_comment_liked(
                    sender=BlogComment.likes.through,
                    instance=comment,
                    action="post_add",
                    reverse=False,
                    pk_set={self.user.id},
                )
            except Exception:
                self.fail("Signal should handle async errors gracefully")



import logging
from typing import Any

from django.conf import settings
from django.db import transaction
from django.db.models import QuerySet
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from djmoney.money import Money

from order.enum.status import OrderStatus
from order.models.item import OrderItem
from order.models.order import Order
from order.signals import order_canceled, order_status_changed

logger = logging.getLogger(__name__)


class OrderServiceError(Exception):
    pass


class ProductNotFoundError(OrderServiceError):
    pass


class InsufficientStockError(OrderServiceError):
    pass


class InvalidOrderDataError(OrderServiceError):
    pass


class OrderNotFoundError(OrderServiceError):
    pass


class InvalidStatusTransitionError(OrderServiceError):
    pass


class OrderCancellationError(OrderServiceError):
    pass


class OrderService:
    @classmethod
    def get_order_by_id(cls, order_id: int) -> Order | None:
        order = (
            Order.objects.select_related("user", "pay_way", "country", "region")
            .prefetch_related("items")
            .get(id=order_id)
        )

        return order

    @classmethod
    def get_order_by_uuid(cls, uuid: str) -> Order | None:
        order = (
            Order.objects.select_related("user", "pay_way", "country", "region")
            .prefetch_related("items")
            .get(uuid=uuid)
        )

        return order

    @classmethod
    @transaction.atomic
    def create_order(
        cls,
        order_data: dict[str, Any],
        items_data: list[dict[str, Any]],
        user=None,
    ) -> Order:
        try:
            if user and user.is_authenticated:
                order_data["user"] = user

            order = Order.objects.create(**order_data)

            target_currency = (
                order.shipping_price.currency
                if order.shipping_price
                else settings.DEFAULT_CURRENCY
            )

            for item_data in items_data:
                product = item_data.get("product")
                quantity = item_data.get("quantity", 0)

                if not product:
                    raise ProductNotFoundError(
                        _("Product is required for order items")
                    )

                if quantity <= 0:
                    raise InvalidOrderDataError(
                        _(
                            "Invalid quantity {quantity} for product {product_id}"
                        ).format(
                            quantity=quantity,
                            product_id=getattr(product, "id", _("unknown")),
                        )
                    )

                if product.stock < quantity:
                    raise InsufficientStockError(
                        _(
                            "Product {product_name} does not have enough stock. Available: {available}, Requested: {requested}"
                        ).format(
                            product_name=product.name or product.id,
                            available=product.stock,
                            requested=quantity,
                        )
                    )

                item_to_create = item_data.copy()

                product_price = product.final_price
                if product_price.currency != target_currency:
                    item_to_create["price"] = Money(
                        product_price.amount, target_currency
                    )
                else:
                    item_to_create["price"] = product_price

                OrderItem.objects.create(order=order, **item_to_create)

            order.paid_amount = order.calculate_order_total_amount()
            order.save(update_fields=["paid_amount"])

            logger.info(
                f"Order {order.id} created successfully with {len(items_data)} items"
            )

            return order

        except (
            ProductNotFoundError,
            InsufficientStockError,
            InvalidOrderDataError,
        ) as e:
            logger.warning(
                f"Re-raising {e.__class__.__name__} as ValueError for backward compatibility"
            )
            raise ValueError(str(e)) from e
        except Exception as e:
            logger.error(
                f"Unexpected error creating order: {e!s}", exc_info=True
            )
            raise InvalidOrderDataError(
                _("Failed to create order: {error}").format(error=str(e))
            ) from e

    @classmethod
    @transaction.atomic
    def update_order_status(cls, order: Order, new_status: str) -> Order:
        try:
            if not new_status:
                raise InvalidStatusTransitionError(
                    _("New status cannot be empty")
                )

            if order.status == new_status:
                logger.info(f"Order {order.id} status is already {new_status}")
                return order

            allowed_transitions = {
                OrderStatus.PENDING: [
                    OrderStatus.PROCESSING,
                    OrderStatus.CANCELED,
                ],
                OrderStatus.PROCESSING: [
                    OrderStatus.SHIPPED,
                    OrderStatus.CANCELED,
                ],
                OrderStatus.SHIPPED: [
                    OrderStatus.DELIVERED,
                    OrderStatus.RETURNED,
                ],
                OrderStatus.DELIVERED: [
                    OrderStatus.COMPLETED,
                    OrderStatus.RETURNED,
                ],
                OrderStatus.CANCELED: [],
                OrderStatus.COMPLETED: [],
                OrderStatus.RETURNED: [OrderStatus.REFUNDED],
                OrderStatus.REFUNDED: [],
            }

            if new_status not in allowed_transitions.get(order.status, []):
                error_message = _(
                    "Cannot transition from {old_status} to {new_status}. Allowed transitions: {allowed_transitions}"
                ).format(
                    old_status=order.status,
                    new_status=new_status,
                    allowed_transitions=allowed_transitions.get(
                        order.status, []
                    ),
                )
                logger.warning(
                    f"Invalid status transition for order {order.id}: {error_message}"
                )
                raise InvalidStatusTransitionError(error_message)

            old_status = order.status

            order.status = new_status
            order.status_updated_at = timezone.now()
            order.save(update_fields=["status", "status_updated_at"])

            order_status_changed.send(
                sender=cls,
                order=order,
                old_status=old_status,
                new_status=new_status,
            )

            logger.info(
                f"Order {order.id} status updated from {old_status} to {new_status}"
            )

            return order

        except InvalidStatusTransitionError as e:
            logger.warning(
                f"Re-raising {e.__class__.__name__} as ValueError for backward compatibility"
            )
            raise ValueError(str(e)) from e

    @classmethod
    def get_user_orders(cls, user_id: int) -> QuerySet:
        user_orders = (
            Order.objects.filter(user_id=user_id)
            .select_related("pay_way", "country", "region")
            .prefetch_related("items", "items__product")
            .order_by("-created_at")
        )

        return user_orders

    @classmethod
    @transaction.atomic
    def cancel_order(cls, order: Order) -> Order:
        if not order.can_be_canceled:
            error_message = _(
                "Order in status {status} cannot be canceled. Only orders in PENDING or PROCESSING status can be canceled."
            ).format(status=order.status)
            logger.warning(f"Cannot cancel order {order.id}: {error_message}")
            raise ValueError(error_message)

        try:
            for item in order.items.select_related("product").all():
                product = item.product
                if hasattr(product, "stock"):
                    logger.info(
                        f"Restoring stock for product {product.id}: +{item.quantity}"
                    )
                    product.stock += item.quantity
                    product.save(update_fields=["stock"])

            old_status = order.status
            order.status = OrderStatus.CANCELED
            order.status_updated_at = timezone.now()
            order.save(update_fields=["status", "status_updated_at"])

            order_canceled.send(
                sender=cls, order=order, previous_status=old_status
            )

            logger.info(
                f"Order {order.id} canceled successfully (previous status: {old_status})"
            )

            return order

        except Exception as e:
            logger.error(
                f"Error canceling order {order.id}: {e!s}", exc_info=True
            )
            raise ValueError(
                _("Failed to cancel order: {error}").format(error=str(e))
            ) from e

    @classmethod
    def calculate_shipping_cost(
        cls,
        order_value: Money,
        country_id: int | None = None,
        region_id: int | None = None,
    ) -> Money:
        from extra_settings.models import Setting  # noqa: PLC0415

        base_shipping_cost = Setting.get(
            "CHECKOUT_SHIPPING_PRICE", default=3.00
        )
        free_shipping_threshold = Setting.get(
            "FREE_SHIPPING_THRESHOLD", default=50.00
        )

        if order_value.amount >= float(free_shipping_threshold):
            return Money(0, order_value.currency)

        shipping_cost = float(base_shipping_cost)

        if country_id:
            from country.models import Country  # noqa: PLC0415

            try:
                country = Country.objects.get(id=country_id)

                if (
                    hasattr(country, "shipping_multiplier")
                    and country.shipping_multiplier
                ):
                    shipping_cost *= country.shipping_multiplier

            except (ImportError, Country.DoesNotExist) as e:
                logger.warning(
                    f"Country with ID {country_id} not found or country module not available: {e!s}"
                )

        if region_id:
            from region.models import Region  # noqa: PLC0415

            try:
                region = Region.objects.get(id=region_id)

                if (
                    hasattr(region, "shipping_adjustment")
                    and region.shipping_adjustment
                ):
                    shipping_cost += region.shipping_adjustment

            except (ImportError, Region.DoesNotExist) as e:
                logger.warning(
                    f"Region with ID {region_id} not found or region module not available: {e!s}"
                )

        return Money(shipping_cost, order_value.currency)



from __future__ import annotations

from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import (
    extend_schema,
    extend_schema_view,
)
from rest_framework import status
from rest_framework.decorators import action

from rest_framework.response import Response

from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet
from core.utils.serializers import (
    create_schema_view_config,
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from product.filters.favourite import ProductFavouriteFilter
from product.models.favourite import ProductFavourite
from product.serializers.favourite import (
    ProductDetailResponseSerializer,
    ProductFavouriteByProductsRequestSerializer,
    ProductFavouriteByProductsResponseSerializer,
    ProductFavouriteDetailSerializer,
    ProductFavouriteSerializer,
    ProductFavouriteWriteSerializer,
)

req_serializers: RequestSerializersConfig = {
    "create": ProductFavouriteWriteSerializer,
    "update": ProductFavouriteWriteSerializer,
    "partial_update": ProductFavouriteWriteSerializer,
    "favourites_by_products": ProductFavouriteByProductsRequestSerializer,
}

res_serializers: ResponseSerializersConfig = {
    "create": ProductFavouriteWriteSerializer,
    "list": ProductFavouriteSerializer,
    "retrieve": ProductFavouriteDetailSerializer,
    "update": ProductFavouriteWriteSerializer,
    "partial_update": ProductFavouriteWriteSerializer,
    "product": ProductDetailResponseSerializer,
    "favourites_by_products": ProductFavouriteByProductsResponseSerializer,
}

schema_config = create_schema_view_config(
    model_class=ProductFavourite,
    display_config={
        "tag": "Product Favourites",
    },
    request_serializers=req_serializers,
    response_serializers=res_serializers,
)


@extend_schema_view(**schema_config)
class ProductFavouriteViewSet(BaseModelViewSet):
    queryset = ProductFavourite.objects.select_related(
        "user", "product"
    ).prefetch_related("product__translations")
    filterset_class = ProductFavouriteFilter
    ordering_fields = [
        "id",
        "user_id",
        "product_id",
        "created_at",
        "updated_at",
    ]
    ordering = ["-created_at"]
    search_fields = [
        "user__username",
        "product__translations__name",
    ]
    response_serializers = res_serializers
    request_serializers = req_serializers

    def create(self, request, *args, **kwargs):
        req_serializer = self.get_request_serializer()
        request_serializer = req_serializer(
            data=request.data, context=self.get_serializer_context()
        )
        request_serializer.is_valid(raise_exception=True)
        request_serializer.save(user=request.user)

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            request_serializer.instance, context=self.get_serializer_context()
        )

        headers = self.get_success_headers(response_serializer.data)
        return Response(
            response_serializer.data,
            status=status.HTTP_201_CREATED,
            headers=headers,
        )

    @extend_schema(
        operation_id="getProductFavouriteProduct",
        summary=_("Get favourite product details"),
        description=_(
            "Get detailed information about the product in this favourite entry."
        ),
        tags=["Product Favourites"],
        responses={
            200: ProductDetailResponseSerializer,
        },
    )
    @action(detail=True, methods=["GET"])
    def product(self, request, *args, **kwargs):
        product_favourite = self.get_object()

        self.ordering_fields = []
        self.ordering = []
        self.search_fields = []

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(
            product_favourite.product, context=self.get_serializer_context()
        )
        return Response(response_serializer.data, status=status.HTTP_200_OK)

    @extend_schema(
        operation_id="getProductFavouritesByProducts",
        summary=_("Get favourites by product IDs"),
        description=_(
            "Get favourite entries for the specified product IDs. Requires authentication."
        ),
        tags=["Product Favourites"],
        request=ProductFavouriteByProductsRequestSerializer,
        responses={
            200: {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "id": {"type": "integer"},
                        "userId": {"type": "integer"},
                        "productId": {"type": "integer"},
                        "createdAt": {"type": "string", "format": "date-time"},
                    },
                    "required": ["id", "userId", "productId", "createdAt"],
                },
            },
            404: ErrorResponseSerializer,
        },
        parameters=[],
    )
    @action(detail=False, methods=["POST"])
    def favourites_by_products(self, request, *args, **kwargs):
        user = request.user
        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(data=request.data)
        request_serializer.is_valid(raise_exception=True)

        product_ids = request_serializer.validated_data["product_ids"]
        favourites = ProductFavourite.objects.filter(
            user=user, product_id__in=product_ids
        ).select_related("user", "product")

        response_serializer_class = self.get_response_serializer()
        response_serializer = response_serializer_class(favourites, many=True)
        return Response(response_serializer.data, status=status.HTTP_200_OK)



from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters

from core.filters.camel_case_filters import CamelCaseTimeStampFilterSet
from core.filters.core import UUIDFilterMixin
from user.models.subscription import SubscriptionTopic, UserSubscription


class SubscriptionTopicFilter(UUIDFilterMixin, CamelCaseTimeStampFilterSet):
    category = filters.ChoiceFilter(
        choices=SubscriptionTopic.TopicCategory.choices,
        help_text=_("Filter by topic category"),
    )
    is_active = filters.BooleanFilter(
        help_text=_("Filter by active status"),
    )
    is_default = filters.BooleanFilter(
        help_text=_("Filter by default subscription status"),
    )
    requires_confirmation = filters.BooleanFilter(
        help_text=_("Filter by confirmation requirement"),
    )

    slug = filters.CharFilter(
        lookup_expr="icontains",
        help_text=_("Filter by slug (partial match)"),
    )
    slug_exact = filters.CharFilter(
        field_name="slug",
        lookup_expr="exact",
        help_text=_("Filter by exact slug"),
    )

    name = filters.CharFilter(
        field_name="translations__name",
        lookup_expr="icontains",
        help_text=_("Filter by name (partial match)"),
    )
    description = filters.CharFilter(
        field_name="translations__description",
        lookup_expr="icontains",
        help_text=_("Filter by description (partial match)"),
    )

    # Custom filter methods
    has_subscribers = filters.BooleanFilter(
        method="filter_has_subscribers",
        help_text=_("Filter topics that have subscribers"),
    )

    class Meta:
        model = SubscriptionTopic
        fields = {
            "id": ["exact", "in"],
            "category": ["exact"],
            "is_active": ["exact"],
            "is_default": ["exact"],
            "requires_confirmation": ["exact"],
            "slug": ["exact", "icontains"],
            "created_at": ["gte", "lte", "date"],
            "updated_at": ["gte", "lte", "date"],
            "uuid": ["exact"],
        }

    def filter_has_subscribers(self, queryset, name, value):
        if value is True:
            return queryset.filter(subscribers__isnull=False).distinct()
        elif value is False:
            return queryset.filter(subscribers__isnull=True)
        return queryset


class UserSubscriptionFilter(UUIDFilterMixin, CamelCaseTimeStampFilterSet):
    topic = filters.ModelChoiceFilter(
        queryset=SubscriptionTopic.objects.filter(is_active=True),
        help_text=_("Filter by subscription topic"),
    )
    status = filters.ChoiceFilter(
        choices=UserSubscription.SubscriptionStatus.choices,
        help_text=_("Filter by subscription status"),
    )

    topic_category = filters.ChoiceFilter(
        field_name="topic__category",
        choices=SubscriptionTopic.TopicCategory.choices,
        help_text=_("Filter by topic category"),
    )

    subscribed_after = filters.DateTimeFilter(
        field_name="subscribed_at",
        lookup_expr="gte",
        help_text=_("Filter subscriptions created after this date"),
    )
    subscribed_before = filters.DateTimeFilter(
        field_name="subscribed_at",
        lookup_expr="lte",
        help_text=_("Filter subscriptions created before this date"),
    )
    unsubscribed_after = filters.DateTimeFilter(
        field_name="unsubscribed_at",
        lookup_expr="gte",
        help_text=_("Filter subscriptions unsubscribed after this date"),
    )
    unsubscribed_before = filters.DateTimeFilter(
        field_name="unsubscribed_at",
        lookup_expr="lte",
        help_text=_("Filter subscriptions unsubscribed before this date"),
    )

    topic_slug = filters.CharFilter(
        field_name="topic__slug",
        lookup_expr="icontains",
        help_text=_("Filter by topic slug (partial match)"),
    )
    topic_slug_exact = filters.CharFilter(
        field_name="topic__slug",
        lookup_expr="exact",
        help_text=_("Filter by exact topic slug"),
    )

    topic_name = filters.CharFilter(
        field_name="topic__translations__name",
        lookup_expr="icontains",
        help_text=_("Filter by topic name (partial match)"),
    )
    topic_description = filters.CharFilter(
        field_name="topic__translations__description",
        lookup_expr="icontains",
        help_text=_("Filter by topic description (partial match)"),
    )

    # Custom filter methods
    is_confirmed = filters.BooleanFilter(
        method="filter_is_confirmed",
        help_text=_("Filter by confirmation status"),
    )
    has_metadata = filters.BooleanFilter(
        method="filter_has_metadata",
        help_text=_("Filter subscriptions that have metadata"),
    )

    class Meta:
        model = UserSubscription
        fields = {
            "id": ["exact", "in"],
            "topic": ["exact"],
            "status": ["exact"],
            "subscribed_at": ["gte", "lte", "date"],
            "unsubscribed_at": ["gte", "lte", "date"],
            "created_at": ["gte", "lte", "date"],
            "updated_at": ["gte", "lte", "date"],
            "uuid": ["exact"],
        }

    def filter_is_confirmed(self, queryset, name, value):
        if value is True:
            return queryset.filter(confirmation_token="")
        elif value is False:
            return queryset.exclude(confirmation_token="")
        return queryset

    def filter_has_metadata(self, queryset, name, value):
        if value is True:
            return queryset.exclude(metadata={})
        elif value is False:
            return queryset.filter(metadata={})
        return queryset
